import holoviews as hv
import pandas as pd
import pytest

from holonote.annotate import Style
from holonote.annotate.display import _default_color
from holonote.tests.util import get_editor, get_indicator


def compare_indicator_color(indicator, style):
    if isinstance(indicator.opts["color"], hv.dim):
        if isinstance(style.color, hv.dim):
            assert str(indicator.opts["color"]) == str(style.color)
        elif style.color is None and style._colormap:
            assert dict(zip(style._groupby[1], _default_color)) == style._colormap
        else:
            expected_dim = hv.dim("__selected__").categorize(
                categories={True: style.selection_color}, default=style.color
            )
            assert str(indicator.opts["color"]) == str(expected_dim)
    else:
        assert indicator.opts["color"] == style.color


def compare_style(cat_annotator):
    style = cat_annotator.style
    indicator = next(get_indicator(cat_annotator, hv.VSpans))
    compare_indicator_color(indicator, style)
    expected_dim = hv.dim("__selected__").categorize(
        categories={True: style.selection_alpha}, default=style.alpha
    )
    assert str(indicator.opts["alpha"]) == str(expected_dim)

    editor = get_editor(cat_annotator, hv.VSpan)
    assert editor.opts["color"] == style.edit_color
    assert editor.opts["alpha"] == style.edit_alpha


def get_selected_indicator_data(annotator) -> pd.Series:
    df = pd.concat([i.data for i in get_indicator(annotator, hv.VLines)])
    return df["__selected__"]


def test_style_accessor(cat_annotator) -> None:
    assert isinstance(cat_annotator.style, Style)


def test_style_default(cat_annotator) -> None:
    compare_style(cat_annotator)


def test_style_default_select(cat_annotator) -> None:
    # Select the first value
    index = cat_annotator.df.index[0]
    cat_annotator.select_by_index(index)
    compare_style(cat_annotator)
    sel_data = get_selected_indicator_data(cat_annotator)
    assert sel_data[index]
    assert sel_data.sum() == 1

    # remove it again
    cat_annotator.select_by_index()
    compare_style(cat_annotator)
    sel_data = get_selected_indicator_data(cat_annotator)
    assert sel_data.sum() == 0


def test_style_change_color_alpha(cat_annotator) -> None:
    style = cat_annotator.style
    style.color = "blue"
    style.alpha = 0.1

    compare_style(cat_annotator)

    style.edit_color = "yellow"
    style.edit_alpha = 1
    cat_annotator.set_regions(x=(0, 1))  # To update plot
    compare_style(cat_annotator)


def test_style_color_dim(cat_annotator):
    style = cat_annotator.style
    style.color = hv.dim("category").categorize(
        categories={"B": "red", "A": "blue", "C": "green"}, default="grey"
    )
    compare_style(cat_annotator)


@pytest.mark.xfail(
    reason="hv.dim is not supported for selection.color in the current implementation"
)
def test_style_selection_color(cat_annotator):
    style = cat_annotator.style
    style.selection_color = "blue"
    compare_style(cat_annotator)
    sel_data = get_selected_indicator_data(cat_annotator)
    assert sel_data.sum() == 0

    # Select the first value
    index = cat_annotator.df.index[0]
    cat_annotator.select_by_index(index)
    compare_style(cat_annotator)
    sel_data = get_selected_indicator_data(cat_annotator)
    assert sel_data[index]
    assert sel_data.sum() == 1

    # remove it again
    cat_annotator.select_by_index()
    compare_style(cat_annotator)
    sel_data = get_selected_indicator_data(cat_annotator)
    assert sel_data.sum() == 0


def test_style_error_color_dim_and_selection(cat_annotator):
    style = cat_annotator.style
    style.color = hv.dim("category").categorize(
        categories={"B": "red", "A": "blue", "C": "green"}, default="grey"
    )
    style.selection_color = "blue"
    msg = "'Style.color' cannot be a `hv.dim` / `None` when 'Style.selection_color' is not None"
    with pytest.raises(ValueError, match=msg):
        compare_style(cat_annotator)


def test_style_opts(cat_annotator):
    cat_annotator.style.opts = {"line_width": 2}
    compare_style(cat_annotator)

    indicator = next(get_indicator(cat_annotator, hv.VSpans))
    assert indicator.opts["line_width"] == 2

    editor = get_editor(cat_annotator, hv.VSpan)
    assert "line_width" not in editor.opts.get().kwargs

    cat_annotator.style.edit_opts = {"line_width": 3}
    cat_annotator.set_regions(x=(0, 1))  # To update plot
    editor = get_editor(cat_annotator, hv.VSpan)
    assert editor.opts["line_width"] == 3


@pytest.mark.parametrize("opts", [{"alpha": 0.5}, {"color": "red"}], ids=["alpha", "color"])
def test_style_opts_warn(cat_annotator, opts):
    msg = "Color and alpha opts should be set directly on the style object"
    with pytest.raises(UserWarning, match=msg):
        cat_annotator.style.opts = opts


def test_style_reset(cat_annotator) -> None:
    style = cat_annotator.style
    style.color = "blue"
    compare_style(cat_annotator)

    style.reset()
    assert style.color == Style.color
    compare_style(cat_annotator)


def test_groupby_color_change(cat_annotator) -> None:
    cat_annotator.groupby = "category"
    cat_annotator.visible = ["A", "B", "C"]

    indicators = hv.render(cat_annotator.get_display("x").static_indicators()).renderers
    color_cycle = cat_annotator.style._colormap.values()
    for indicator, expected_color in zip(indicators, color_cycle):
        assert indicator.glyph.fill_color == expected_color



from __future__ import annotations

import holoviews as hv
import pytest
from holoviews.operation.datashader import rasterize

from holonote.tests.util import get_editor, get_indicator


def get_editor_data(annotator, element_type, kdims=None):
    el = get_editor(annotator, element_type, kdims)
    return getattr(el, "data", None)


def get_indicator_data(annotator, element_type, kdims=None):
    for el in get_indicator(annotator, element_type, kdims):
        yield el.data


def test_set_regions_range1d(annotator_range1d) -> None:
    annotator = annotator_range1d
    annotator.get_display("TIME")

    # No regions has been set
    output = get_editor_data(annotator, hv.VSpan)
    expected = [None, None]
    assert output == expected

    # Setting regions
    annotator.set_regions(TIME=(-0.25, 0.25))
    output = get_editor_data(annotator, hv.VSpan)
    expected = [-0.25, 0.25]
    assert output == expected

    # Adding annotation and remove selection regions.
    annotator.add_annotation(description="Test")
    output = get_editor_data(annotator, hv.VSpan)
    expected = [None, None]
    assert output == expected

    output = next(get_indicator_data(annotator, hv.Rectangles))
    output1 = output.iloc[0][["start[TIME]", "end[TIME]"]].tolist()
    expected1 = [-0.25, 0.25]
    assert output1 == expected1
    output2 = output.iloc[0]["description"]
    expected2 = "Test"
    assert output2 == expected2


def test_edit_streams(annotator_range1d) -> None:
    annotator = annotator_range1d
    edit_streams = annotator.get_display("TIME").edit_streams
    assert len(edit_streams) == 1
    assert "box_select" in edit_streams
    assert "tap" not in edit_streams
    assert "lasso_select" not in edit_streams


def test_set_regions_range2d(annotator_range2d) -> None:
    annotator = annotator_range2d
    annotator.get_display("x", "y")

    # No regions has been set
    output = get_editor_data(annotator, hv.Rectangles)
    assert output.empty

    # Setting regions
    annotator.set_regions(x=(-0.25, 0.25), y=(-0.25, 0.25))
    output = get_editor_data(annotator, hv.Rectangles).iloc[0].to_list()
    expected = [-0.25, -0.25, 0.25, 0.25]
    assert output == expected

    # Adding annotation and remove selection regions.
    annotator.add_annotation(description="Test")
    output = get_editor_data(annotator, hv.Rectangles)
    assert output.empty

    output = next(get_indicator_data(annotator, hv.Rectangles))
    output1 = output.iloc[0][["start[x]", "start[y]", "end[x]", "end[y]"]].tolist()
    expected1 = [-0.25, -0.25, 0.25, 0.25]
    assert output1 == expected1
    output2 = output.iloc[0]["description"]
    expected2 = "Test"
    assert output2 == expected2


def test_set_regions_multiple(multiple_annotators):
    annotator = multiple_annotators
    annotator.get_display("TIME")
    annotator.get_display("x", "y")

    # No regions has been set
    # Time annotation
    output = get_editor_data(annotator, hv.VSpan, "TIME")
    expected = [None, None]
    assert output == expected
    # xy annotation
    output = get_editor_data(annotator, hv.Rectangles, ("x", "y"))
    assert output.empty

    # Setting regions
    annotator.set_regions(TIME=(-0.25, 0.25), x=(-0.25, 0.25), y=(-0.25, 0.25))
    # Time annotation
    output = get_editor_data(annotator, hv.VSpan, "TIME")
    expected = [-0.25, 0.25]
    assert output == expected
    # xy annotation
    output = get_editor_data(annotator, hv.Rectangles, ("x", "y")).iloc[0].to_list()
    expected = [-0.25, -0.25, 0.25, 0.25]
    assert output == expected

    # Adding annotation and remove selection regions.
    annotator.add_annotation(description="Test")
    # Time annotation
    output = get_editor_data(annotator, hv.VSpan, "TIME")
    expected = [None, None]
    assert output == expected

    output = next(get_indicator_data(annotator, hv.Rectangles, "TIME"))
    output1 = output.iloc[0][["start[TIME]", "end[TIME]"]].tolist()
    expected1 = [-0.25, 0.25]
    assert output1 == expected1
    output2 = output.iloc[0]["description"]
    expected2 = "Test"
    assert output2 == expected2

    # xy annotation
    output = get_editor_data(annotator, hv.Rectangles, ("x", "y"))
    assert output.empty

    output = next(get_indicator_data(annotator, hv.Rectangles, ("x", "y")))
    output1 = output.iloc[0][["start[x]", "start[y]", "end[x]", "end[y]"]].tolist()
    expected1 = [-0.25, -0.25, 0.25, 0.25]
    assert output1 == expected1
    output2 = output.iloc[0]["description"]
    expected2 = "Test"
    assert output2 == expected2


def test_editable_enabled(annotator_range1d):
    annotator_range1d.get_display("TIME")
    assert annotator_range1d._displays

    annotator_range1d.editable_enabled = False
    for display in annotator_range1d._displays.values():
        assert not display.editable_enabled

    annotator_range1d.editable_enabled = True
    for display in annotator_range1d._displays.values():
        assert display.editable_enabled


def test_selection_enabled(annotator_range1d, element_range1d):
    annotator_range1d.get_display("TIME")
    assert annotator_range1d._displays

    annotator_range1d.selection_enabled = False
    for display in annotator_range1d._displays.values():
        assert not display.selection_enabled

    annotator_range1d.selection_enabled = True
    for display in annotator_range1d._displays.values():
        assert display.selection_enabled


def test_groupby(cat_annotator):
    cat_annotator.groupby = "category"
    iter_indicator = get_indicator(cat_annotator, hv.VSpans)
    indicator = next(iter_indicator)
    assert indicator.data.shape == (2, 5)
    assert (indicator.data["category"] == "A").all()

    indicator = next(iter_indicator)
    assert indicator.data.shape == (2, 5)
    assert (indicator.data["category"] == "B").all()

    indicator = next(iter_indicator)
    assert indicator.data.shape == (1, 5)
    assert (indicator.data["category"] == "C").all()

    with pytest.raises(StopIteration):
        next(iter_indicator)


def test_groupby_visible(cat_annotator):
    cat_annotator.groupby = "category"
    cat_annotator.visible = ["A"]
    iter_indicator = get_indicator(cat_annotator, hv.VSpans)
    indicator = next(iter_indicator)
    assert indicator.data.shape == (2, 5)
    assert (indicator.data["category"] == "A").all()

    with pytest.raises(StopIteration):
        next(iter_indicator)


def test_multiply_overlay(annotator_range1d):
    el1 = hv.Curve([], kdims=["TIME"])
    el2 = hv.Curve([], kdims=["TIME"])
    el = el1 * el2
    assert isinstance(el, hv.Overlay)

    el = el * annotator_range1d
    hv.render(el)
    assert isinstance(el.last, hv.Overlay)

    el.opts(width=100)
    hv.render(el)
    assert el.last.opts.get().kwargs.get("width") == 100


def test_multiply_ndoverlay(annotator_range1d):
    el1 = hv.Curve([], kdims=["TIME"])
    el2 = hv.Curve([], kdims=["TIME"])
    el = hv.NdOverlay({1: el1, 2: el2})
    assert isinstance(el, hv.NdOverlay)

    el = el * annotator_range1d
    hv.render(el)
    assert isinstance(el.last, hv.Overlay)

    el.opts(width=100)
    hv.render(el)
    assert el.last.opts.get().kwargs.get("width") == 100


def test_multiply_layout(annotator_range1d):
    el1 = hv.Curve([], kdims=["TIME"])
    el2 = hv.Curve([], kdims=["TIME"])
    el = el1 + el2
    assert isinstance(el, hv.Layout)

    el = el * annotator_range1d
    hv.render(el)
    assert isinstance(el, hv.Layout)

    el.opts(width=100)
    hv.render(el)
    assert el.opts.get().kwargs.get("width") == 100


def test_multiply_dynamicmap(annotator_range1d):
    # rasterize creates a dynamicmap
    el1 = rasterize(hv.Curve([], kdims=["TIME"]))
    assert isinstance(el1, hv.DynamicMap)

    el = el1 * annotator_range1d
    hv.render(el)
    # an overlay including a dynamicmap is still a dynamicmap
    assert isinstance(el, hv.DynamicMap)


def test_multiply_dynamicmap_layout(annotator_range1d):
    el1 = rasterize(hv.Curve([], kdims=["TIME"]))
    el2 = hv.Curve([], kdims=["TIME"])
    el = el1 + el2
    assert isinstance(el, hv.Layout)

    el = el * annotator_range1d
    hv.render(el)
    assert isinstance(el, hv.Layout)



import hashlib
import json
import os
from glob import glob


def calculate_sha256(file_path):
    sha256_hash = hashlib.sha256()

    with open(file_path, "rb") as file:
        for byte_block in iter(lambda: file.read(4096), b""):
            sha256_hash.update(byte_block)

    return sha256_hash.hexdigest()


with open("package.json") as f:
    package = json.load(f)
pyodide_version = package["dependencies"]["pyodide"].removeprefix("^")

path = "pyodide-lock.json"
url = f"https://cdn.jsdelivr.net/pyodide/v{pyodide_version}/full"

with open(path) as f:
    data = json.load(f)

for p in data["packages"].values():
    if not p["file_name"].startswith("http"):
        p["file_name"] = os.path.join(url, p["file_name"])


# Special handling of holonote
whl_file = glob("../../dist/*.whl")[0]
hn = data["packages"]["holonote"]
hn["version"] = os.environ["VERSION"]
hn["file_name"] = os.path.basename(whl_file)
hn["sha256"] = calculate_sha256(whl_file)
hn["imports"] = ["holonote"]  # Not completely sure why this is empty

# To avoid importing it in the notebooks, we can't add it to pandas directly
# as fastparquet depends on it. So we add it to hvplot instead.
data["packages"]["hvplot"]["depends"].extend(["fastparquet"])

data["packages"]["holoviews"]["depends"].extend(["pyparsing"])


with open(path, "w") as f:
    data = json.dump(data, f)



from __future__ import annotations

import contextlib
import datetime as dt
from typing import TYPE_CHECKING, Any

import panel as pn
import param
from packaging.version import Version
from panel.viewable import Viewer

from ..annotate.display import _default_color

if TYPE_CHECKING:
    from holonote.annotate import Annotator

PN13 = Version(pn.__version__) >= Version("1.3.0")


class PanelWidgets(Viewer):
    reset_on_apply = param.Boolean(default=True, doc="Reset fields widgets on apply")

    mapping = {
        str: pn.widgets.TextInput,
        bool: pn.widgets.Checkbox,
        dt.datetime: pn.widgets.DatePicker,
        dt.date: pn.widgets.DatePicker,
        int: pn.widgets.IntSlider,
        float: pn.widgets.FloatSlider,
    }

    def __init__(
        self,
        annotator: Annotator,
        field_values: dict[str, Any] | None = None,
        as_popup: bool = False,
        **params,
    ):
        super().__init__(**params)
        self._layouts = {}
        self.annotator = annotator
        self.annotator.snapshot()
        self._widget_mode_group = pn.widgets.RadioButtonGroup(
            name="Mode", options=["+", "-", "✏"], width=90
        )
        self._widget_apply_button = pn.widgets.Button(name="✓", width=20)
        self._widget_revert_button = pn.widgets.Button(name="↺", width=20)
        self._widget_commit_button = pn.widgets.Button(name="▲", width=20)
        if PN13:
            self._add_button_description()

        if field_values is None:
            self._fields_values = {k: "" for k in self.annotator.fields}
        else:
            self._fields_values = {k: field_values.get(k, "") for k in self.annotator.fields}
        self._fields_widgets = self._create_fields_widgets(self._fields_values)
        self._create_visible_widget()

        self._set_standard_callbacks()

        self._layout = pn.Column(self.fields_widgets, self.tool_widgets)
        if self.visible_widget is not None:
            self._layout.insert(0, self.visible_widget)

        self._as_popup = as_popup
        if self._as_popup:
            self._layout.visible = False
            displays = self.annotator._displays
            if not displays:
                kdims = list(self.annotator.spec.keys())
                display = self.annotator.get_display(*kdims)
                display.indicators()
            for display in displays.values():
                if display.region_format in ("range", "range-range"):
                    stream = display._edit_streams[0]
                elif display.region_format in ("point", "point-point"):
                    stream = display._edit_streams[1]
                self._register_stream_popup(stream)
                self._register_tap_popup(display)
                self._register_double_tap_clear(display)

    def _create_visible_widget(self):
        if self.annotator.groupby is None:
            self.visible_widget = None
            return
        style = self.annotator.style
        if style.color is None and style._colormap is None:
            data = sorted(self.annotator.df[self.annotator.groupby].unique())
            colormap = dict(zip(data, _default_color))
        else:
            colormap = style._colormap
        if isinstance(colormap, dict):
            stylesheet = """
            option:after {
              content: "";
              width: 10px;
              height: 10px;
              position: absolute;
              border-radius: 50%;
              left: calc(100% - var(--design-unit, 4) * 2px - 3px);
              top: 20%;
              border: 1px solid black;
              opacity: 0.5;
            }"""
            for i, color in enumerate(colormap.values()):
                stylesheet += f"""
            option:nth-child({i + 1}):after {{
                background-color: {color};
            }}"""
        else:
            stylesheet = ""

        options = list(colormap)
        self.visible_widget = pn.widgets.MultiSelect(
            name="Visible",
            options=options,
            value=self.annotator.visible or options,
            stylesheets=[stylesheet],
        )
        self.annotator.visible = self.visible_widget

    def _add_button_description(self):
        from bokeh.models import Tooltip
        from bokeh.models.dom import HTML

        html_table = """
        <h3>Mode:</h3>
        <table style="width:100%">
        <tr>
            <td style="min-width:50px">+</td>
            <td>Add annotation</td>
        </tr>
        <tr>
            <td>-</td>
            <td>Delete annotation</td>
        </tr>
        <tr>
            <td>✏</td>
            <td>Edit annotation</td>
        </tr>
        </table>
        """

        self._widget_mode_group.description = Tooltip(
            content=HTML(html_table),
            position="bottom",
        )
        self._widget_apply_button.description = Tooltip(content="Apply changes", position="bottom")
        self._widget_revert_button.description = Tooltip(
            content="Revert changes not yet saved to database", position="bottom"
        )
        self._widget_commit_button.description = Tooltip(
            content="Save to database", position="bottom"
        )

    @property
    def tool_widgets(self):
        return pn.Row(
            self._widget_apply_button,
            pn.Spacer(width=10),
            self._widget_mode_group,
            pn.Spacer(width=10),
            self._widget_revert_button,
            self._widget_commit_button,
        )

    def _create_fields_widgets(self, fields_values):
        fields_widgets = {}
        for widget_name, default in fields_values.items():
            if isinstance(default, param.Parameter):
                parameterized = type("widgets", (param.Parameterized,), {widget_name: default})
                pane = pn.Param(parameterized)
                fields_widgets[widget_name] = pane.layout[1]
            elif isinstance(default, list):
                fields_widgets[widget_name] = pn.widgets.Select(
                    value=default[0], options=default, name=widget_name
                )
            else:
                widget_type = self.mapping[type(default)]
                if issubclass(widget_type, pn.widgets.TextInput):
                    fields_widgets[widget_name] = widget_type(
                        value=default, placeholder=widget_name, name=widget_name
                    )
                else:
                    fields_widgets[widget_name] = widget_type(value=default, name=widget_name)
        return fields_widgets

    @property
    def fields_widgets(self):
        accordion = False  # Experimental
        widgets = pn.Column(*self._fields_widgets.values())
        if accordion:
            return pn.Accordion(("fields", widgets))
        else:
            return widgets

    def _reset_fields_widgets(self):
        for widget_name, default in self._fields_values.items():
            if isinstance(default, param.Parameter):
                default = default.default
            if isinstance(default, list):
                default = default[0]
            with contextlib.suppress(Exception):
                # TODO: Fix when lists (for categories, not the same as the default!)
                self._fields_widgets[widget_name].value = default

    def _callback_apply(self, event):
        selected_ind = (
            self.annotator.selected_indices[0]
            if len(self.annotator.selected_indices) == 1
            else None
        )
        self.annotator.select_by_index()

        if self._widget_mode_group.value in ["+", "✏"]:
            fields_values = {k: v.value for k, v in self._fields_widgets.items()}
            if self._widget_mode_group.value == "+":
                self.annotator.add_annotation(**fields_values)
                if self.reset_on_apply:
                    self._reset_fields_widgets()
            elif (self._widget_mode_group.value == "✏") and (selected_ind is not None):
                self.annotator.update_annotation_fields(
                    selected_ind, **fields_values
                )  # TODO: Handle only changed
        elif self._widget_mode_group.value == "-" and selected_ind is not None:
            self.annotator.delete_annotation(selected_ind)

    def _get_layout(self, name):
        def close_layout(event):
            layout.visible = False

        layout = self._layouts.get(name)
        if not layout:
            layout = self._layout.clone(visible=False)
            self._widget_apply_button.on_click(close_layout)
            self._layouts[name] = layout
        return layout

    def _hide_layouts(self):
        for layout in self._layouts.values():
            layout.visible = False

    def _register_stream_popup(self, stream):
        def _popup(*args, **kwargs):
            layout = self._get_layout(stream.name)
            with param.parameterized.batch_call_watchers(self):
                self._hide_layouts()
                self._widget_mode_group.value = "+"
                layout.visible = True
                return layout

        stream.popup = _popup

    def _register_tap_popup(self, display):
        def tap_popup(x, y) -> None:  # Tap tool must be enabled on the element
            layout = self._get_layout("tap")
            if self.annotator.selection_enabled:
                with param.parameterized.batch_call_watchers(self):
                    self._hide_layouts()
                    layout.visible = True
                    return layout

        display._tap_stream.popup = tap_popup

    def _register_double_tap_clear(self, display):
        def double_tap_toggle(x, y):
            layout = self._get_layout("doubletap")
            if layout.visible:
                with param.parameterized.batch_call_watchers(self):
                    self._hide_layouts()
                    layout.visible = True
                    return layout

        try:
            tools = display._element.opts["tools"]
        except KeyError:
            tools = []
        display._element.opts(tools=[*tools, "doubletap"])
        display._double_tap_stream.popup = double_tap_toggle

    def _callback_commit(self, event):
        self.annotator.commit()

    def _watcher_selected_indices(self, event):
        if len(event.new) != 1:
            return
        selected_index = event.new[0]
        # if self._widget_mode_group.value == '✏':
        for name, widget in self._fields_widgets.items():
            value = self.annotator.annotation_table._field_df.loc[selected_index][name]
            widget.value = value

    def _watcher_mode_group(self, event):
        with param.parameterized.batch_call_watchers(self):
            if event.new in ("-", "✏"):
                self.annotator.selection_enabled = True
            elif event.new == "+":
                self.annotator.editable_enabled = True
                self.annotator.selection_enabled = False

            for widget in self._fields_widgets.values():
                widget.disabled = event.new == "-"

    def _set_standard_callbacks(self):
        self._widget_apply_button.on_click(self._callback_apply)
        self._widget_revert_button.on_click(lambda event: self.annotator.revert_to_snapshot())
        self._widget_commit_button.on_click(self._callback_commit)
        self.annotator.param.watch(self._watcher_selected_indices, "selected_indices")
        self._widget_mode_group.param.watch(self._watcher_mode_group, "value")

    def __panel__(self):
        return self._layout.clone(visible=True)



import numpy as np
import pandas as pd
import pytest

from holonote.annotate import Annotator, SQLiteDB

pytestmark = pytest.mark.benchmark


def setup_annotator(items, filename=":memory:") -> tuple[Annotator, pd.DataFrame]:
    rng = np.random.default_rng(1337)
    start_time = np.arange(0, items * 2, 2)
    end_time = start_time + 1
    data = pd.DataFrame(
        {
            "start_time": start_time,
            "end_time": end_time,
            "description": rng.choice(["A", "B"], items),
        }
    )
    annotator = Annotator(
        {"TIME": int}, fields=["description"], connector=SQLiteDB(filename=filename)
    )
    return annotator, data


@pytest.mark.parametrize("items", [10, 100, 1000])
def test_define_annotations(benchmark, items) -> None:
    @benchmark
    def bench():
        annotator, data = setup_annotator(items)
        annotator.define_annotations(data, TIME=("start_time", "end_time"))


@pytest.mark.parametrize("items", [10, 100, 1000])
def test_commit(benchmark, items, tmp_path) -> None:
    @benchmark
    def bench():
        annotator, data = setup_annotator(items, filename=str(tmp_path / "test.db"))
        annotator.define_annotations(data, TIME=("start_time", "end_time"))
        annotator.commit()



from __future__ import annotations

import contextlib
from collections.abc import Iterator

import holoviews as hv
import numpy as np
import pandas as pd
import pytest

from holonote.annotate import Annotator, SQLiteDB, UUIDHexStringKey

optional_markers = {
    "benchmark": {
        "help": "Run benchmarks",
        "marker-descr": "Benchmark test marker",
        "skip-reason": "Test only runs with the --benchmark option.",
    },
}


def pytest_addoption(parser):
    for marker, info in optional_markers.items():
        parser.addoption(f"--{marker}", action="store_true", default=False, help=info["help"])


def pytest_configure(config):
    for marker, info in optional_markers.items():
        config.addinivalue_line("markers", "{}: {}".format(marker, info["marker-descr"]))


def pytest_collection_modifyitems(config, items):
    skipped, selected = [], []
    markers = [m for m in optional_markers if config.getoption(f"--{m}")]
    empty = not markers
    for item in items:
        if empty and any(m in item.keywords for m in optional_markers):
            skipped.append(item)
        elif empty or (not empty and any(m in item.keywords for m in markers)):
            selected.append(item)
        else:
            skipped.append(item)

    config.hook.pytest_deselected(items=skipped)
    items[:] = selected


@pytest.fixture()
def conn_sqlite_uuid(tmp_path) -> Iterator[SQLiteDB]:
    conn = SQLiteDB(filename=str(tmp_path / "test.db"), primary_key=UUIDHexStringKey())
    yield conn
    with contextlib.suppress(Exception):
        conn.cursor.close()
    with contextlib.suppress(Exception):
        conn.con.close()


@pytest.fixture()
def annotator_range1d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"TIME": np.datetime64},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def annotator_point1d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"TIME": (np.datetime64, "point")},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def annotator_range2d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"x": float, "y": float},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def annotator_point2d(conn_sqlite_uuid) -> Annotator:
    anno = Annotator(
        {"x": (float, "point"), "y": (float, "point")},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def multiple_region_annotator(conn_sqlite_uuid) -> Annotator:
    return Annotator(
        {"TIME": np.datetime64},
        fields=["description"],
        # region_types=["Point", "Range"],
        connector=conn_sqlite_uuid,
    )


@pytest.fixture()
def multiple_annotators(conn_sqlite_uuid) -> Annotator:
    return Annotator(
        {"TIME": np.datetime64, "x": float, "y": float},
        fields=["description"],
        connector=conn_sqlite_uuid,
    )


@pytest.fixture()
def multiple_fields_annotator(conn_sqlite_uuid) -> Annotator:
    conn_sqlite_uuid.fields = ["field1", "field2"]
    anno = Annotator(
        {"TIME": np.datetime64},
        fields=["field1", "field2"],
        connector=conn_sqlite_uuid,
    )
    return anno


@pytest.fixture()
def element_range1d() -> hv.Curve:
    time = pd.date_range("2020-01-01", "2020-01-10", freq="D").to_numpy()
    return hv.Curve(time, kdims=["TIME"])


@pytest.fixture()
def element_range2d() -> hv.Image:
    x = np.arange(10)
    xy = x[:, np.newaxis] * x
    return hv.Image(xy, kdims=["x", "y"])


@pytest.fixture()
def cat_annotator(conn_sqlite_uuid) -> Annotator:
    # Initialize annotator
    annotator = Annotator(
        {"x": float},
        fields=["description", "category"],
        connector=conn_sqlite_uuid,
        groupby="category",
    )
    # Add data to annotator
    data = {
        "category": ["A", "B", "A", "C", "B"],
        "start_number": [1, 6, 11, 16, 21],
        "end_number": [5, 10, 15, 20, 25],
        "description": list("ABCDE"),
    }
    annotator.define_annotations(pd.DataFrame(data), x=("start_number", "end_number"))
    # Setup display
    annotator.get_display("x")
    return annotator



from __future__ import annotations

import panel as pn

from holonote.app import PanelWidgets


def test_panel_app(annotator_range1d):
    w = PanelWidgets(annotator_range1d)
    assert isinstance(w.fields_widgets, pn.Column)
    assert isinstance(w.tool_widgets, pn.Row)


def test_as_popup(annotator_range1d):
    w = PanelWidgets(annotator_range1d, as_popup=True)
    assert not w._layout.visible
    for display in w.annotator._displays.values():
        assert display._edit_streams[0].popup
        assert display._tap_stream.popup
    assert w.__panel__().visible



from .annotator import Annotator
from .connector import (
    AutoIncrementKey,
    Connector,
    SQLiteDB,
    UUIDBinaryKey,
    UUIDHexStringKey,
)
from .display import Style
from .table import AnnotationTable

__all__ = (
    "AnnotationTable",
    "Annotator",
    "AutoIncrementKey",
    "Connector",
    "SQLiteDB",
    "Style",
    "UUIDBinaryKey",
    "UUIDHexStringKey",
)



import holoviews as hv

bk_renderer = hv.renderer("bokeh")


def _get_display(annotator, kdims=None) -> hv.Element:
    if kdims is None:
        kdims = next(iter(annotator._displays))
    kdims = (kdims,) if isinstance(kdims, str) else tuple(kdims)
    disp = annotator.get_display(*kdims)
    bk_renderer.get_plot(disp.element)  # Trigger rendering
    return disp


def get_editor(annotator, element_type, kdims=None):
    el = _get_display(annotator, kdims).region_editor()
    for e in el.last.traverse():
        if isinstance(e, element_type):
            return e


def get_indicator(annotator, element_type, kdims=None):
    si = _get_display(annotator, kdims).indicators().last
    yield from si.data.values()



from pathlib import Path

PATH = (Path(__file__).parents[1] / "doc" / "tutorial").resolve(strict=True)

for rst in PATH.glob("*.rst"):
    lines = rst.open().readlines()

    for idx, line in enumerate(lines):  # noqa: B007
        if line.startswith("*****"):
            break

    new_lines = [*lines[: idx - 1], *lines[idx + 3 :]]
    rst.open("w").writelines(new_lines)



import os
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from uuid import uuid4

import pytest


def main() -> None:
    example_path = Path(__file__).parents[1] / "examples"

    with TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        os.symlink(example_path / ".." / "pyproject.toml", tmpdir / "pyproject.toml")

        for file in sorted(example_path.rglob("*.ipynb")):
            if file.parent.name == ".ipynb_checkpoints":
                continue
            test_folder = tmpdir / uuid4().hex[:8]
            test_file = test_folder / file.relative_to(example_path)
            test_file.parent.mkdir(parents=True, exist_ok=True)
            os.symlink(file, test_file)
            os.symlink(example_path / "assets", test_folder / "assets")

        os.chdir(tmpdir)  # To not have large paths in pytest output
        exit_code = pytest.main(["--nbval-lax"])
        os.chdir(example_path)  # To be able to clean up folder on Windows

    sys.exit(exit_code)


if __name__ == "__main__":
    main()



from __future__ import annotations

import holoviews as hv
import numpy as np
import pandas as pd
import pytest

from holonote.annotate import Annotator, SQLiteDB


@pytest.mark.skip("Not supported having multiple region types for same dimension")
def test_multipoint_range_commit_insertion(multiple_region_annotator):
    descriptions = ["A point insertion", "A range insertion"]
    timestamp = np.datetime64("2022-06-06")
    multiple_region_annotator.set_point(timestamp)
    multiple_region_annotator.add_annotation(description=descriptions[0])

    start, end = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
    multiple_region_annotator.set_range(start, end)
    multiple_region_annotator.add_annotation(description=descriptions[1])

    multiple_region_annotator.commit(return_commits=True)

    # FIXME! Index order is inverted?
    df = pd.DataFrame(
        {
            "uuid": pd.Series(multiple_region_annotator.df.index[::-1], dtype=object),
            "point_TIME": [timestamp, pd.NaT],
            "start_TIME": [pd.NaT, start],
            "end_TIME": [pd.NaT, end],
            "description": descriptions,
        }
    ).set_index("uuid")

    sql_df = multiple_region_annotator.connector.load_dataframe()
    pd.testing.assert_frame_equal(sql_df, df)


def test_infer_kdim_dtype_image():
    xvals, yvals = np.linspace(-4, 0, 202), np.linspace(4, 0, 202)
    xs, ys = np.meshgrid(xvals, yvals)
    image = hv.Image(np.sin(ys * xs), kdims=["A", "B"])
    assert Annotator._infer_kdim_dtypes(image) == {"A": np.float64, "B": np.float64}


def test_infer_kdim_dtype_curve():
    curve = hv.Curve(
        (np.arange("2005-02", "2005-03", dtype="datetime64[D]"), range(28)), kdims=["TIME"]
    )
    assert Annotator._infer_kdim_dtypes(curve) == {"TIME": np.datetime64}


def test_multiplot_add_annotation(multiple_annotators):
    start_time, end_time = np.datetime64("2005-02-13"), np.datetime64("2005-02-16")
    multiple_annotators.set_regions(TIME=(start_time, end_time))
    multiple_annotators.set_regions(x=(-0.25, 0.25), y=(-0.1, 0.1))
    multiple_annotators.add_annotation(description="Multi-plot annotation", uuid="A")
    multiple_annotators.commit()

    d = {
        "start[TIME]": {"A": start_time},
        "end[TIME]": {"A": end_time},
        "start[x]": {"A": -0.25},
        "end[x]": {"A": 0.25},
        "start[y]": {"A": -0.1},
        "end[y]": {"A": 0.1},
        "description": {"A": "Multi-plot annotation"},
    }
    expected = pd.DataFrame(d)
    expected.index.name = "uuid"
    pd.testing.assert_frame_equal(multiple_annotators.df, expected)


class TestAnnotatorMultipleStringFields:
    def test_insertion_values(self, multiple_fields_annotator):
        start, end = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
        multiple_fields_annotator.set_regions(TIME=(start, end))
        multiple_fields_annotator.add_annotation(
            field1="A test field", field2="Another test field"
        )
        commits = multiple_fields_annotator.commit(return_commits=True)
        kwargs = commits[0]["kwargs"]
        assert len(commits) == 1, "Only one insertion commit made"
        assert "uuid" in kwargs, "Expected uuid primary key in kwargs"
        kwargs.pop("uuid")
        assert kwargs == {
            "field1": "A test field",
            "field2": "Another test field",
            "start_TIME": start,
            "end_TIME": end,
        }

    def test_commit_insertion(self, multiple_fields_annotator):
        start, end = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
        field1 = "A test field"
        field2 = "Another test field"
        multiple_fields_annotator.set_regions(TIME=(start, end))
        multiple_fields_annotator.add_annotation(field1=field1, field2=field2)
        multiple_fields_annotator.commit(return_commits=True)

        df = pd.DataFrame(
            {
                "uuid": pd.Series(multiple_fields_annotator.df.index[0], dtype=object),
                "start_TIME": [start],
                "end_TIME": [end],
                "field1": [field1],
                "field2": [field2],
            }
        ).set_index("uuid")

        sql_df = multiple_fields_annotator.connector.load_dataframe()
        pd.testing.assert_frame_equal(sql_df, df)

    def test_commit_update_set_range(self, multiple_fields_annotator):
        start1, end1 = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
        start2, end2 = np.datetime64("2023-06-06"), np.datetime64("2023-06-08")
        multiple_fields_annotator.set_regions(TIME=(start1, end1))
        multiple_fields_annotator.add_annotation(field1="Field 1.1", field2="Field 1.2")
        multiple_fields_annotator.set_regions(TIME=(start2, end2))
        multiple_fields_annotator.add_annotation(field1="Field 2.1", field2="Field 2.2")
        multiple_fields_annotator.commit(return_commits=True)
        multiple_fields_annotator.update_annotation_fields(
            multiple_fields_annotator.df.index[0], field1="NEW Field 1.1"
        )
        multiple_fields_annotator.commit(return_commits=True)
        sql_df = multiple_fields_annotator.connector.load_dataframe()
        assert set(sql_df["field1"]) == {"NEW Field 1.1", "Field 2.1"}

    def test_commit_update_set_region(self, multiple_fields_annotator):
        start1, end1 = np.datetime64("2022-06-06"), np.datetime64("2022-06-08")
        start2, end2 = np.datetime64("2023-06-06"), np.datetime64("2023-06-08")
        multiple_fields_annotator.set_regions(TIME=(start1, end1))
        multiple_fields_annotator.add_annotation(field1="Field 1.1", field2="Field 1.2")
        multiple_fields_annotator.set_regions(TIME=(start2, end2))
        multiple_fields_annotator.add_annotation(field1="Field 2.1", field2="Field 2.2")
        multiple_fields_annotator.commit(return_commits=True)
        multiple_fields_annotator.update_annotation_fields(
            multiple_fields_annotator.df.index[0], field1="NEW Field 1.1"
        )
        multiple_fields_annotator.commit(return_commits=True)
        sql_df = multiple_fields_annotator.connector.load_dataframe()
        assert set(sql_df["field1"]) == {"NEW Field 1.1", "Field 2.1"}


@pytest.mark.parametrize("method", ["new", "same"])
def test_reconnect(method, tmp_path):
    db_path = str(tmp_path / "test.db")

    if method == "new":
        conn1 = SQLiteDB(filename=db_path)
        conn2 = SQLiteDB(filename=db_path)
    elif method == "same":
        conn1 = conn2 = SQLiteDB(filename=db_path)

    # Create annotator with data and commit
    a1 = Annotator(
        spec={"TIME": np.datetime64},
        fields=["description"],
        connector=conn1,
    )
    times = pd.date_range("2022-06-09", "2022-06-13")
    for t1, t2 in zip(times[:-1], times[1:]):
        a1.set_regions(TIME=(t1, t2))
        a1.add_annotation(description="A programmatically defined annotation")
    a1.commit(return_commits=True)

    # Save internal dataframes
    a1_df = a1.df.copy()
    a1_region = a1.annotation_table._region_df.copy()
    a1_field = a1.annotation_table._field_df.copy()

    # Add new connector
    a2 = Annotator(
        spec={"TIME": np.datetime64},
        fields=["description"],
        connector=conn2,
    )
    a2_df = a2.df.copy()
    a2_region = a2.annotation_table._region_df.copy()
    a2_field = a2.annotation_table._field_df.copy()

    pd.testing.assert_frame_equal(a1_df, a2_df)
    pd.testing.assert_frame_equal(a1_region, a2_region)
    pd.testing.assert_frame_equal(a1_field, a2_field)


def test_define_annotations_multiple_without_without_all_fields(conn_sqlite_uuid) -> None:
    annotator = Annotator(
        {"TIME": float}, fields=["description", "category"], connector=conn_sqlite_uuid
    )

    # Sample data without description field
    data = {
        "category": ["A", "B", "A", "C", "B"],
        "start_number": [1, 6, 11, 16, 21],
        "end_number": [5, 10, 15, 20, 25],
    }

    annotator.define_annotations(pd.DataFrame(data), TIME=("start_number", "end_number"))

    # Add description field
    data["description"] = (None,) * 5

    annotator.define_annotations(pd.DataFrame(data), TIME=("start_number", "end_number"))


def test_define_annotations_with_different_field(conn_sqlite_uuid) -> None:
    annotator = Annotator({"TIME": float}, fields=["description"], connector=conn_sqlite_uuid)
    data = {
        "category": ["A", "B", "A", "C", "B"],
        "start_number": [1, 6, 11, 16, 21],
        "end_number": [5, 10, 15, 20, 25],
    }
    annotator.define_annotations(
        pd.DataFrame(data), TIME=("start_number", "end_number"), description="category"
    )
    assert (annotator.df["description"] == data["category"]).all()


def test_define_annotations_with_multiple_field_warn(conn_sqlite_uuid) -> None:
    annotator = Annotator(
        {"TIME": float}, fields=["description", "category"], connector=conn_sqlite_uuid
    )
    data = {
        "category": ["A", "B", "A", "C", "B"],
        "start_number": [1, 6, 11, 16, 21],
        "end_number": [5, 10, 15, 20, 25],
    }

    # Description set to use category will give a warning,
    # which raises an error in the testsuite
    msg = "Input 'category' has overlapping name with a field or spec"
    with pytest.raises(UserWarning, match=msg):
        annotator.define_annotations(
            pd.DataFrame(data), TIME=("start_number", "end_number"), description="category"
        )


def test_define_annotations_clear_region(conn_sqlite_uuid) -> None:
    # See: https://github.com/holoviz/holonote/pull/50
    annotator = Annotator({"x": float}, connector=conn_sqlite_uuid)
    data = {"description": ["A", "B"], "start_number": [0.1, 0.5], "end_number": [0.4, 0.8]}
    annotator.define_annotations(pd.DataFrame(data), x=("start_number", "end_number"))
    assert annotator.region == {}
    assert annotator._last_region == {}


def test_update_region(multiple_annotators, conn_sqlite_uuid) -> None:
    annotator = multiple_annotators

    # Add one annotation and commit
    first = pd.Timestamp("2020-01-01"), pd.Timestamp("2020-01-02")
    annotator.set_regions(TIME=first)
    annotator.add_annotation(description="test")
    annotator.commit()

    # In the Annotation Table
    df = annotator.df
    assert df.shape[0] == 1
    assert (df["start[TIME]"] == first[0]).all()
    assert (df["end[TIME]"] == first[1]).all()

    # In the database
    df = conn_sqlite_uuid.load_dataframe()
    assert df.shape[0] == 1
    assert (df["start_TIME"] == first[0]).all()
    assert (df["end_TIME"] == first[1]).all()

    # Update annotation region and commit
    second = pd.Timestamp("2022-01-01"), pd.Timestamp("2022-01-02")
    annotator.set_regions(TIME=second)
    annotator.update_annotation_region(df.index[0])
    annotator.commit()

    # In the Annotation Table
    df = annotator.df
    assert df.shape[0] == 1
    assert (df["start[TIME]"] == second[0]).all()
    assert (df["end[TIME]"] == second[1]).all()

    # In the database
    df = conn_sqlite_uuid.load_dataframe()
    assert df.shape[0] == 1
    assert (df["start_TIME"] == second[0]).all()
    assert (df["end_TIME"] == second[1]).all()


class TestEvent:
    @pytest.fixture(autouse=True)
    def _setup_count(self, multiple_annotators):
        self.count = {"create": 0, "update": 0, "delete": 0}

        def count(event) -> None:
            self.count[event.type] += 1

        multiple_annotators.on_event(count)

    def check(self, create=0, update=0, delete=0):
        assert self.count["create"] == create
        assert self.count["update"] == update
        assert self.count["delete"] == delete

    def test_create(self, multiple_annotators):
        annotator = multiple_annotators
        annotator.set_regions(x=(0, 1))
        self.check(create=0, update=0, delete=0)
        annotator.add_annotation(description="test")
        self.check(create=1, update=0, delete=0)
        annotator.commit()

    def test_update_fields(self, multiple_annotators):
        annotator = multiple_annotators
        annotator.set_regions(x=(0, 1))
        self.check(create=0, update=0, delete=0)
        annotator.add_annotation(description="test")
        self.check(create=1, update=0, delete=0)

        index = annotator.df.index[0]
        annotator.update_annotation_fields(index, description="updated")
        self.check(create=1, update=1, delete=0)

    def test_update_regions(self, multiple_annotators):
        annotator = multiple_annotators
        annotator.set_regions(x=(0, 1))
        self.check(create=0, update=0, delete=0)
        annotator.add_annotation(description="test")
        self.check(create=1, update=0, delete=0)

        index = annotator.df.index[0]
        annotator.set_regions(x=(1, 2))
        annotator.update_annotation_region(index)
        self.check(create=1, update=1, delete=0)

    def test_delete(self, multiple_annotators):
        annotator = multiple_annotators
        annotator.set_regions(x=(0, 1))
        self.check(create=0, update=0, delete=0)
        annotator.add_annotation(description="test")
        self.check(create=1, update=0, delete=0)

        index = annotator.df.index[0]
        annotator.delete_annotation(index)
        self.check(create=1, update=0, delete=1)



from __future__ import annotations

import datetime as dt
import hashlib
import os
import sqlite3
import sys
import uuid
from contextlib import contextmanager
from functools import cache
from pathlib import Path
from shutil import copyfile
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import param

if TYPE_CHECKING:
    from .typing import SpecDict


@cache
def _sqlite_adapters() -> None:
    # Most of the following code has been copied here from in Python 3.11:
    # `sqlite3.dbapi2.register_adapters_and_converters`.
    # Including minor modifications to source code to pass linting
    # https://docs.python.org/3/license.html#psf-license
    # Extra adapters have been added for dt.time, numpy.datetime64, and pd.Timestamp.

    def adapt_date(val):
        return val.isoformat()

    def adapt_datetime(val):
        return val.isoformat(" ")

    def adapt_time(t) -> str:
        return f"{t.hour:02d}:{t.minute:02d}:{t.second:02d}.{t.microsecond:06d}"

    def adapt_np_datetime64(val):
        return np.datetime_as_string(val, unit="us").replace("T", " ")

    def convert_date(val):
        return dt.date(*map(int, val.split(b"-")))

    def convert_timestamp(val):
        datepart, timepart = val.split(b" ")
        year, month, day = map(int, datepart.split(b"-"))
        timepart_full = timepart.split(b".")
        hours, minutes, seconds = map(int, timepart_full[0].split(b":"))
        microseconds = int(f"{timepart_full[1].decode():0<6.6}") if len(timepart_full) == 2 else 0

        val = dt.datetime(year, month, day, hours, minutes, seconds, microseconds)
        return val

    if sys.version_info >= (3, 12):
        # Python 3.12 has removed datetime support from sqlite3
        # https://github.com/python/cpython/pull/93095
        sqlite3.register_adapter(dt.date, adapt_date)
        sqlite3.register_adapter(dt.datetime, adapt_datetime)
        sqlite3.register_converter("date", convert_date)
        sqlite3.register_converter("timestamp", convert_timestamp)

    sqlite3.register_adapter(dt.time, adapt_time)
    sqlite3.register_adapter(np.datetime64, adapt_np_datetime64)
    sqlite3.register_adapter(pd.Timestamp, adapt_datetime)


def _get_valid_sqlite_name(name: object):
    # See https://stackoverflow.com/questions/6514274/how-do-you-escape-strings\
    # -for-sqlite-table-column-names-in-python
    # Ensure the string can be encoded as UTF-8.
    # Ensure the string does not include any NUL characters.
    # Replace all " with "".
    # Wrap the entire thing in double quotes.
    # Inspired from pandas.io.sql._get_valid_sqlite_name

    try:
        uname = str(name).encode("utf-8", "strict").decode("utf-8")
    except UnicodeError as err:
        msg = f"Cannot convert identifier to UTF-8: '{name}'"
        raise ValueError(msg) from err

    if not len(uname):
        msg = "Empty table or column name specified"
        raise ValueError(msg)

    nul_index = uname.find("\x00")
    if nul_index >= 0:
        msg = "SQLite identifier cannot contain NULs"
        raise ValueError(msg)
    return '"' + uname.replace('"', '""') + '"'


class PrimaryKey(param.Parameterized):
    """
    Generator of the primary key used to keep track of annotations in
    HoloViews.

    The generated key is used to reference annotations until they are
    committed, at which point they may 1) be inserted in the database as
    the primary key value (policy='insert') 2) are checked against the
    primary key value chosen by the database which is expected to match
    in most cases.

    In real situations where the key is chosen by the database, the key
    generated will *not* always match the actual key assigned. The
    policy parameter decides the resulting behavior in these cases.
    """

    field_name = param.String(default="id", allow_None=False)

    policy = param.ObjectSelector(
        default="ignore-mismatch",
        objects=["insert", "ignore-mismatch", "warn-mismatch", "error-mismatch"],
    )

    schema = param.String(default="INTEGER PRIMARY KEY", constant=True, allow_None=False)

    connector_class = param.String(default="SQLiteDB")

    def __call__(self, connector, key_list=None):
        """
        The key list is the current list of index values that are
        outstanding (i.e. have not been committed).
        """
        raise NotImplementedError

    def validate(self, database_key, local_key):
        if self.policy in ["insert", "ignore-mismatch"]:
            return

        if database_key != local_key:
            print("MISMATCH")

    def cast(self, value):
        "Cast a user supplied value to a known supported type"
        raise NotImplementedError


class AutoIncrementKey(PrimaryKey):
    """
    AUTOINCREMENT needed to prevent reuse of ids from deleted rows:
    https://www.sqlite.org/autoinc.html
    """

    field_name = param.String(default="id", allow_None=False)

    policy = param.ObjectSelector(
        default="ignore-mismatch",
        objects=["insert", "ignore-mismatch", "warn-mismatch", "error-mismatch"],
    )

    schema = param.String(
        default="INTEGER PRIMARY KEY AUTOINCREMENT", constant=True, allow_None=False
    )

    def __call__(self, connector, key_list=None):
        key_list_max = max(key_list) if key_list else 0
        connector_max = connector.max_rowid()
        connector_max = 0 if connector_max is None else connector_max
        max_rowid = max(key_list_max, connector_max)
        return max_rowid + 1

    def cast(self, value):
        return int(value)  # e.g. np.int64 from a pandas index won't work!


class UUIDHexStringKey(PrimaryKey):  # Probably the better default
    """
    Example of 'insert' policy where the generated primary key value can
    be inserted in the database as uuids are independent and not
    expected to clash.
    """

    field_name = param.String(default="uuid", allow_None=False)

    policy = param.ObjectSelector(
        default="insert", objects=["insert", "ignore-mismatch", "warn-mismatch", "error-mismatch"]
    )

    schema = param.String("TEXT PRIMARY KEY", constant=True, allow_None=False)

    length = param.Integer(default=32, bounds=(4, 32))

    def __call__(self, connector, key_list=None):
        return uuid.uuid4().hex[: self.length]

    def cast(self, value):
        return str(value)


class UUIDBinaryKey(PrimaryKey):
    """
    Example of 'insert' policy where the generated primary key value can
    be inserted in the database as uuids are independent and not
    expected to clash.
    """

    field_name = param.String(default="uuid", allow_None=False)

    policy = param.ObjectSelector(
        default="insert", objects=["insert", "ignore-mismatch", "warn-mismatch", "error-mismatch"]
    )

    schema = param.String("BINARY PRIMARY KEY", constant=True, allow_None=False)

    def __call__(self, connector, key_list=None):
        return uuid.uuid4().bytes

    def cast(self, value):
        return bytes(value)


class WidgetKey(PrimaryKey):
    """
    Placeholder for a concept where the user can insert a primary key
    value via a widget.
    """


class Connector(param.Parameterized):
    """
    Base class that support the auto-generated default schema
    """

    primary_key = param.Parameter(default=UUIDHexStringKey(), allow_None=False)

    commit_hook = param.Parameter(default=None, doc="Callback, applies default schema if None")

    fields = param.List(default=None, doc="List of column names for domain-specific fields")

    transforms = param.Dict(
        default={
            "insert": lambda x: x,
            "update": lambda x: x,
            "delete": lambda x: x,
            "save": lambda x: x,
            "load": lambda x: x,
        }
    )

    operation_mapping = {}  # Mapping from operation type to corresponding connector method

    # iterate on all the possible types
    type_mapping = {
        bool: "BOOLEAN",
        str: "TEXT",
        float: "REAL",
        int: "INTEGER",
        np.datetime64: "TIMESTAMP",
        dt.date: "TIMESTAMP",
        dt.datetime: "TIMESTAMP",
        dt.time: "TIME",
        pd.Timedelta: "INTEGER",
        pd.Timestamp: "TIMESTAMP",
        np.dtype("datetime64[ns]"): "TIMESTAMP",
        np.dtype("<M8"): "TIMESTAMP",
        np.float64: "REAL",
        np.float32: "REAL",
        np.float16: "REAL",
        np.int8: "INTEGER",
        np.int16: "INTEGER",
        np.int32: "INTEGER",
        np.int64: "INTEGER",
        np.uint8: "INTEGER",
        np.uint16: "INTEGER",
        np.uint32: "INTEGER",
    }

    @classmethod
    def field_value_to_type(cls, value):
        if isinstance(value, list):
            assert all(isinstance(el, str) for el in value), "Only string enums supported"
            return str
        elif hasattr(value, "dtype"):
            return value.dtype
        elif type(value) in cls.type_mapping:
            return type(value)
        elif isinstance(value, param.Parameter) and value.default is not None:
            return type(value.default)
        else:
            msg = f"Connector cannot handle type {type(value)!s}"
            raise TypeError(msg)

    @classmethod
    def schema_from_field_values(cls, fields):
        "Given a dictionary of fields fields and values return the field schemas"
        return {
            name: cls.type_mapping[cls.field_value_to_type(value)]
            for name, value in fields.items()
        }

    @classmethod
    def schema_from_field_types(cls, fields):
        "Given a dictionary of fields fields and values return the field schemas"
        return {name: cls.type_mapping[value] for name, value in fields.items()}

    @classmethod
    def expand_region_column_schema(cls, region_types, kdim_dtypes):
        region_type_column_prefixes = {
            "Range": ("start", "end"),
            "Point": ("point",),
            "point": ("point",),
            "range": ("start", "end"),
        }
        fields = {}
        for region_type in region_types:
            for column in region_type_column_prefixes[region_type]:
                for kdim, kdim_dtype in kdim_dtypes.items():
                    fields[f"{column}_{kdim}"] = cls.type_mapping[kdim_dtype]
        return fields

    @classmethod
    def generate_schema(cls, primary_key, all_region_types, all_kdim_dtypes, field_types):
        schemas = {primary_key.field_name: primary_key.schema}
        region_schemas = {}
        for region_types, kdim_dtypes in zip(all_region_types, all_kdim_dtypes):
            region_schemas = dict(
                region_schemas, **cls.expand_region_column_schema(region_types, kdim_dtypes)
            )
        schemas = dict(schemas, **region_schemas)
        return dict(schemas, **cls.schema_from_field_types(field_types))

    def _incompatible_schema_check(self, expected_keys, columns, fields, region_type):
        msg_prefix = (
            "Unable to read annotations that were stored with a "
            "schema inconsistent with the current settings:"
        )
        msg_suffix = f"Columns found: {columns}"

        missing_field_columns = set(fields) - set(columns)
        if missing_field_columns:
            msg = f"{msg_prefix} Missing field columns {missing_field_columns}. {msg_suffix}"
            raise Exception(msg)

        non_field_columns = set(columns) - set(fields)
        missing_region_columns = set(expected_keys) - non_field_columns
        if missing_region_columns:
            msg = f"{msg_prefix} Missing {region_type!r} region columns {missing_region_columns}. {msg_suffix}"
            raise Exception(msg)

    def _create_column_schema(self, spec: SpecDict, fields: list[str]) -> None:
        field_dtypes = {col: str for col in fields}  # FIXME - generalize
        all_region_types = [{v["region"] for v in spec.values()}]
        all_kdim_dtypes = [{k: v["type"] for k, v in spec.items()}]
        schema = self.generate_schema(
            self.primary_key, all_region_types, all_kdim_dtypes, field_dtypes
        )
        self.column_schema = schema


class _SQLiteDB(Connector):
    """
    Simple example of a Connector without dependencies, using sqlite3.

    column_schema is a dictionary from column name to SQL CREATE TABLE type declaration.
    """

    filename = param.String(default="annotations.db")

    table_name = param.String(
        default=None,
        allow_None=True,
        doc="""
        The SQL table name to connect to in the database. If None, an
        automatically generated table name will be used.""",
    )

    column_schema = param.Dict(default={})

    operation_mapping = {
        "insert": "add_row",
        "delete": "delete_row",
        "save": "add_rows",
        "update": "update_row",
    }

    _tablename_prefix = "annotations"

    def __init__(self, column_schema=None, connect=True, **params):
        """
        First key in column_schema is assumed to the primary key field if not explicitly specified.
        """
        if column_schema is None:
            column_schema = {}

        params["column_schema"] = column_schema
        self.con = None
        super().__init__(**params)

        if connect:
            self._initialize(column_schema, create_table=False)

    @contextmanager
    def run_transaction(self):
        cur = self.con.cursor()
        try:
            yield cur
            self.con.commit()
        except Exception:
            self.con.rollback()
            raise
        finally:
            cur.close()

    def execute(self, query, *args) -> None:
        with self.run_transaction() as cursor:
            cursor.execute(query, *args)

    def close(self):
        self.con.close()
        self.con = None

    def _initialize(self, column_schema, create_table=True):
        _sqlite_adapters()
        if self.con is None:
            self.con = self._create_database_connection()
        if create_table:
            if self.table_name is None:
                self.table_name = self._generate_table_name(column_schema)
            self.create_table(column_schema=column_schema)

    def _create_database_connection(self) -> sqlite3.Connection:
        return sqlite3.connect(
            self.filename, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES
        )

    @property
    def _safe_table_name(self) -> str:
        return _get_valid_sqlite_name(self.table_name)

    def _generate_table_name(self, column_schema) -> str:
        "Given the column_schema outputs a deterministic table name"
        h = hashlib.new("md5")
        for item in sorted(column_schema):
            h.update(item.encode("utf-8"))
            h.update(column_schema[item].encode("utf-8"))
        return f"{self._tablename_prefix}_{h.hexdigest()[:8]}"

    @property
    def uninitialized(self):
        if self.con is not None:
            return self.table_name not in self.get_tables()
        return True

    def max_rowid(self):
        with self.run_transaction() as cursor:
            return cursor.execute(f"SELECT max(ROWID) from {self._safe_table_name}").fetchone()[0]

    def initialize(self, column_schema):
        self.column_schema = column_schema
        self._initialize(column_schema)

    def load_dataframe(self):
        raw_df = pd.read_sql_query(f"SELECT * FROM {self._safe_table_name}", self.con)
        return raw_df.set_index(self.primary_key.field_name)

    def get_tables(self):
        with self.run_transaction() as cursor:
            res = cursor.execute("SELECT name FROM sqlite_master").fetchmany()
        return [el[0] for el in res]

    def create_table(self, column_schema=None):
        column_schema = column_schema if column_schema else self.column_schema
        column_spec = ",\n".join(
            [f"{_get_valid_sqlite_name(name)} {spec}" for name, spec in column_schema.items()]
        )
        query = f"CREATE TABLE IF NOT EXISTS {self._safe_table_name} ({column_spec});"
        self.execute(query)

    def delete_table(self):
        self.execute(f"DROP TABLE IF EXISTS {self._safe_table_name}")

    def add_rows(self, field_list):  # Used execute_many
        for field in field_list:
            self.add_row(**field)

    def add_row(self, **fields):
        columns, parameters = zip(*fields.items())

        if self.primary_key.policy != "insert":
            columns = columns[1:]
            parameters = parameters[1:]

        placeholders = ", ".join(["?"] * len(parameters))
        column_str = ", ".join(map(_get_valid_sqlite_name, columns))
        query = f"INSERT INTO {self._safe_table_name} ({column_str}) VALUES({placeholders});"
        with self.run_transaction() as cursor:
            cursor.execute(query, parameters)
            self.primary_key.validate(cursor.lastrowid, fields[self.primary_key.field_name])

    def delete_all_rows(self):
        "Obviously a destructive operation!"
        self.execute(f"DELETE FROM {self._safe_table_name};")

    def delete_row(self, id_val):
        self.execute(
            f"DELETE FROM {self._safe_table_name} WHERE {self.primary_key.field_name} = ?",
            (self.primary_key.cast(id_val),),
        )

    def update_row(self, **updates):  # updates as a dictionary OR remove posarg?
        assert self.primary_key.field_name in updates
        id_val = updates.pop(self.primary_key.field_name)
        set_updates = ", ".join('"' + k + '"' + " = ?" for k in updates)
        query = f'UPDATE {self._safe_table_name} SET {set_updates} WHERE "{self.primary_key.field_name}" = ?;'
        self.execute(query, [*updates.values(), id_val])


class _SQLiteDBJupyterLite(_SQLiteDB):
    # Sqlite don't work in JupyterLite environment:
    # https://github.com/jupyterlite/pyodide-kernel/issues/35

    def __init__(self, column_schema=None, connect=True, **params) -> None:
        self.tmp_folder = Path("/tmp/holonote") / os.getcwd().removeprefix("/drive")
        self.tmp_folder.mkdir(parents=True, exist_ok=True)
        self.tmp_file = self.tmp_folder / "annotations.db"

        if os.path.exists(self.filename):
            copyfile(self.filename, self.tmp_file)

        super().__init__(column_schema, connect, **params)

    def _create_database_connection(self) -> sqlite3.Connection:
        class CopyConnection(sqlite3.Connection):
            def commit(_self):
                super().commit()
                copyfile(self.tmp_file, self.filename)

        con = sqlite3.connect(
            self.tmp_file,
            factory=CopyConnection,
            detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
        )
        return con


if "_pyodide" in sys.modules:
    js = __import__("js")

    # JupyterLite does not have document
    SQLiteDB = _SQLiteDB if hasattr(js, "document") else _SQLiteDBJupyterLite

else:

    class SQLiteDB(_SQLiteDB):
        pass



import os

from . import annotate, editor

# Define '__version__'
try:
    if os.path.exists(os.path.join(os.path.dirname(__file__), "..", ".git")):
        # If setuptools_scm is installed (e.g. in a development environment with
        # an editable install), then use it to determine the version dynamically.
        from setuptools_scm import get_version

        # This will fail with LookupError if the package is not installed in
        # editable mode or if Git is not installed.
        __version__ = get_version(root="..", relative_to=__file__)
    else:
        raise FileNotFoundError
except Exception:
    # As a fallback, use the version that is hard-coded in the file.
    try:
        from ._version import __version__
    except ModuleNotFoundError:
        # The user is probably trying to run this without having installed
        # the package.
        __version__ = "0.0.0+unknown"

__all__ = ("__version__", "annotate", "editor")



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest


@pytest.mark.parametrize(
    "define_kwargs",
    [
        {},
        {"TIME": "time", "description": "description"},
        {"TIME": "time"},
        {"TIME": "time", "description": "description", "index": True},
        {"TIME": "time", "index": True},
    ],
)
def test_define_annotations_range1d(annotator_range1d, define_kwargs):
    annotator = annotator_range1d
    annotator_table = annotator.annotation_table
    assert annotator_table._region_df.empty
    assert annotator_table._field_df.empty

    # Create data
    time = pd.date_range("2020-01-01", periods=3, freq="D")
    descriptions = ["Ann0", "Ann1", "Ann2"]
    if define_kwargs:
        data = pd.DataFrame({"time": time, "description": descriptions})
    else:
        data = pd.DataFrame({"TIME": time, "description": descriptions})

    # Define annotations
    annotator.define_annotations(data, **define_kwargs)

    # Validate it
    rdf, fdf = annotator_table._region_df, annotator_table._field_df
    assert (rdf["value"].to_list() == time).all()
    assert (rdf["dim"] == "TIME").all()
    assert (rdf["region"] == "range").all()
    assert fdf["description"].to_list() == ["Ann0", "Ann1", "Ann2"]
    assert fdf.index.unique().size == 3
    assert (rdf["_id"] == fdf.index).all()

    if "index" in define_kwargs:
        assert (fdf.index == ["0", "1", "2"]).all()
    else:
        assert (fdf.index != ["0", "1", "2"]).all()


@pytest.mark.parametrize(
    "define_kwargs",
    [
        {"x": ("start", "end"), "description": "description"},
        {"x": ("start", "end")},
        {"x": ("start", "end"), "description": "description", "index": True},
        {"x": ("start", "end"), "index": True},
    ],
)
def test_define_annotations_range2d(annotator_range2d, define_kwargs):
    annotator = annotator_range2d
    annotator_table = annotator.annotation_table
    assert annotator_table._region_df.empty
    assert annotator_table._field_df.empty

    # Create data
    starts = np.arange(3)
    ends = starts + 2
    descriptions = ["Ann0", "Ann1", "Ann2"]
    data = pd.DataFrame({"start": starts, "end": ends, "description": descriptions})

    # Define annotations
    annotator.define_annotations(data, **define_kwargs)

    # Validate it
    rdf, fdf = annotator_table._region_df, annotator_table._field_df
    assert rdf["value"].to_list() == [(0, 2), (1, 3), (2, 4)]
    assert (rdf["dim"] == "x").all()
    assert (rdf["region"] == "range").all()
    assert fdf["description"].to_list() == ["Ann0", "Ann1", "Ann2"]
    assert fdf.index.unique().size == 3
    assert (rdf["_id"] == fdf.index).all()

    if "index" in define_kwargs:
        assert (fdf.index == ["0", "1", "2"]).all()
    else:
        assert (fdf.index != ["0", "1", "2"]).all()


def test_delete_selected_annotation(annotator_range2d):
    annotator = annotator_range2d
    annotator.set_regions(x=(0, 1))
    annotator.add_annotation(description="Ann0", uuid="0")

    annotator.selected_indices = ["0"]
    assert annotator.selected_indices == ["0"]
    assert annotator.selected_index == "0"

    annotator.delete_annotation("0")
    assert not annotator.selected_indices
    assert annotator.selected_index is None

    msg = "Annotation with index '0' does not exist"
    with pytest.raises(ValueError, match=msg):
        annotator.delete_annotation("0")



import os
from pathlib import Path

import param
from nbsite.shared_conf import *  # noqa: F403

import holonote as hn

param.parameterized.docstring_signature = False
param.parameterized.docstring_describe_params = False


project = "HoloNote"
authors = "HoloNote contributors"
copyright_years["start_year"] = "2023"
copyright = copyright_fmt.format(**copyright_years)
description = "Tools to create, edit and persist annotated regions for HoloViews"
version = release = base_version(hn.__version__)
is_dev = any(ext in version for ext in ("a", "b", "rc"))

HOLONOTE_ROOT = Path(hn.__file__).parents[1]
# Remove database files if they exist
for db in (HOLONOTE_ROOT / "examples").rglob("*.db"):
    db.unlink()


# For the interactivity warning box created by nbsite to point to the right
# git tag instead of the default i.e. main.
os.environ["BRANCH"] = f"v{release}"

html_static_path += ["_static"]

html_css_files += ["css/custom.css"]

html_theme = "pydata_sphinx_theme"
# html_favicon = "_static/icons/favicon.ico"

html_theme_options = {
    # "logo": {
    #     "image_light": "_static/logo_horizontal_light_theme.png",
    #     "image_dark": "_static/logo_horizontal_dark_theme.png",
    # },
    "github_url": "https://github.com/holoviz/holonote",
    "icon_links": [
        {
            "name": "Discourse",
            "url": "https://discourse.holoviz.org/",
            "icon": "fa-brands fa-discourse",
        },
        {
            "name": "Discord",
            "url": "https://discord.gg/UXdtYyGVQX",
            "icon": "fa-brands fa-discord",
        },
    ],
    "pygment_light_style": "material",
    "pygment_dark_style": "material",
    "header_links_before_dropdown": 5,
    "secondary_sidebar_items": [
        "github-stars-button",
        "jupyterlitelink",
        "page-toc",
    ],
}

extensions += ["sphinx.ext.napoleon", "nbsite.gallery"]
napoleon_numpy_docstring = True

myst_enable_extensions = ["colon_fence", "deflist"]

jupyterlite_url = "https://holoviz-dev.github.io/holonote-jupyterlite"

nbsite_gallery_conf = {
    "github_org": "holoviz",
    "github_project": "holonote",
    "jupyterlite_url": jupyterlite_url,
    "galleries": {},
}

templates_path += ["_templates"]

html_context.update(
    {
        "last_release": f"v{release}",
        "github_user": "holoviz",
        "github_repo": "holonote",
        "default_mode": "light",
        "jupyterlite_endpoint": jupyterlite_url,
    }
)

nbbuild_patterns_to_take_along = ["simple.html", "*.json", "json_*"]

# Override the Sphinx default title that appends `documentation`
html_title = f"{project} v{version}"


# Patching GridItemCardDirective to be able to substitute the domain name
# in the link option.
# from sphinx_design.cards import CardDirective
# from sphinx_design.grids import GridItemCardDirective

# orig_grid_run = GridItemCardDirective.run


# def patched_grid_run(self):
#     app = self.state.document.settings.env.app
#     existing_link = self.options.get("link")
#     domain = getattr(app.config, "grid_item_link_domain", None)
#     if self.has_content:
#         self.content.replace("|gallery-endpoint|", domain)
#     if existing_link and domain:
#         new_link = existing_link.replace("|gallery-endpoint|", domain)
#         self.options["link"] = new_link
#     return list(orig_grid_run(self))


# GridItemCardDirective.run = patched_grid_run

# orig_card_run = CardDirective.run


# def patched_card_run(self):
#     app = self.state.document.settings.env.app
#     existing_link = self.options.get("link")
#     domain = getattr(app.config, "grid_item_link_domain", None)
#     if existing_link and domain:
#         new_link = existing_link.replace("|gallery-endpoint|", domain)
#         self.options["link"] = new_link
#     return orig_card_run(self)


# CardDirective.run = patched_card_run


def setup(app) -> None:
    try:
        from nbsite.paramdoc import param_formatter, param_skip

        app.connect("autodoc-process-docstring", param_formatter)
        app.connect("autodoc-skip-member", param_skip)
    except ImportError:
        print("no param_formatter (no param?)")

    nbbuild.setup(app)
    app.add_config_value("grid_item_link_domain", "", "html")



from __future__ import annotations

import weakref
from functools import cache, reduce
from typing import TYPE_CHECKING, Any

import colorcet as cc
import holoviews as hv
import numpy as np
import param
from bokeh.models.tools import BoxSelectTool, HoverTool, Tool

from .._warnings import warn

if TYPE_CHECKING:
    from .annotator import Annotator


class _StyleOpts(param.Dict):
    def _validate(self, val) -> None:
        super()._validate(val)
        for k in val:
            if k in ("color", "alpha"):
                warn("Color and alpha opts should be set directly on the style object.")


_default_opts = {"apply_ranges": False, "show_legend": False}

# Make red the first color
_default_color = cc.palette["glasbey_category10"].copy()
_default_color[:4] = [_default_color[3], *_default_color[:3]]


@cache
def _valid_element_opts():
    if not hv.extension._loaded:
        hv.extension("bokeh")
    return hv.opts._element_keywords("bokeh")


class Style(param.Parameterized):
    """
    Style class for controlling the appearance of the annotations
    indicator and editor.

    This can be accessed as an accessor on an annotator object,
    the following will set the annotation color to red:

    >>> from holonote.annotate import Annotator
    >>> annotator = Annotator(...)
    >>> annotator.style.color = "red"

    This will update existing annotation displays and any new
    displays with the new style.

    The style object can also be used to control the appearance
    of the editor and selected indicator:

    >>> annotator.style.edit_color = "blue"
    >>> annotator.style.edit_alpha = 0.5
    >>> annotator.style.selection_color = "green"
    >>> annotator.style.selection_alpha = 0.5

    See the [styling notebook](../../examples/styling.ipynb) for more examples
    of how to use the style object.
    """

    alpha = param.Number(
        default=0.2, bounds=(0, 1), allow_refs=True, doc="Alpha value for non-selected regions"
    )

    selection_alpha = param.Number(
        default=0.7, bounds=(0, 1), allow_refs=True, doc="Alpha value for selected regions"
    )

    edit_alpha = param.Number(
        default=0.4, bounds=(0, 1), allow_refs=True, doc="Alpha value for editing regions"
    )

    color = param.Parameter(default=None, doc="Color of the indicator", allow_refs=True)
    edit_color = param.Parameter(default="blue", doc="Color of the editor", allow_refs=True)
    selection_color = param.Parameter(
        default=None, doc="Color of selection, by the default the same as color", allow_refs=True
    )

    # Indicator opts (default and selection)
    opts = _StyleOpts(default={})
    line_opts = _StyleOpts(default={})
    span_opts = _StyleOpts(default={})
    rectangle_opts = _StyleOpts(default={})

    # Editor opts
    edit_opts = _StyleOpts(default={"line_color": "black"})
    edit_line_opts = _StyleOpts(default={})
    edit_span_opts = _StyleOpts(default={})
    edit_rectangle_opts = _StyleOpts(default={})

    _groupby = ()
    _colormap = None

    @property
    def _color(self):
        if self.color is None:
            if self._groupby:
                # This is the main point of this method
                # we use this to be able to memorize the colormap
                # so the color are the same no matter what
                # the order of the groupby is
                dim = hv.dim(self._groupby[0])
                self._colormap = dict(zip(self._groupby[1], _default_color))
                return dim.categorize(self._colormap)
            else:
                return _default_color[0]

        return self.color

    @property
    def _indicator_selection(self) -> dict[str, tuple]:
        select = {"alpha": (self.selection_alpha, self.alpha)}
        if self.selection_color is not None:
            if isinstance(self._color, hv.dim):
                msg = "'Style.color' cannot be a `hv.dim` / `None` when 'Style.selection_color' is not None"
                raise ValueError(msg)
            else:
                select["color"] = (self.selection_color, self._color)
        return select

    def indicator(self, **select_opts) -> tuple[hv.Options, ...]:
        opts = {**_default_opts, "color": self._color, **select_opts, **self.opts}
        return (
            hv.opts.Rectangles(**opts, **self.rectangle_opts),
            hv.opts.VSpans(**opts, **self.span_opts),
            hv.opts.HSpans(**opts, **self.span_opts),
            hv.opts.VLines(**opts, **self.line_opts),
            hv.opts.HLines(**opts, **self.line_opts),
        )

    def editor(self) -> tuple[hv.Options, ...]:
        opts = {
            **_default_opts,
            "alpha": self.edit_alpha,
            "color": self.edit_color,
            **self.edit_opts,
        }
        return (
            hv.opts.Rectangles(**opts, **self.edit_rectangle_opts),
            hv.opts.VSpan(**opts, **self.edit_span_opts),
            hv.opts.HSpan(**opts, **self.edit_span_opts),
            hv.opts.VLine(**opts, **self.edit_line_opts),
            hv.opts.HLine(**opts, **self.edit_line_opts),
        )

    def reset(self) -> None:
        self._colormap = None
        params = self.param.objects().items()
        self.param.update(**{k: v.default for k, v in params if k != "name"})


class Indicator:
    """
    Collection of class methods that express annotation data as final
    displayed (vectorized) HoloViews object.
    """

    @classmethod
    def points_1d(
        cls, data, region_labels, fields_labels, invert_axes=False, groupby: str | None = None
    ):
        "Vectorizes point regions to VLines. Note does not support hover info"
        vdims = [*fields_labels, "__selected__"]
        element = hv.VLines(data, kdims=region_labels, vdims=vdims)
        hover = cls._build_hover_tool(data)
        return element.opts(tools=[hover])

    @classmethod
    def points_2d(
        cls, data, region_labels, fields_labels, invert_axes=False, groupby: str | None = None
    ):
        "Vectorizes point regions to VLines * HLines. Note does not support hover info"
        msg = "2D point regions not supported yet"
        raise NotImplementedError(msg)
        vdims = [*fields_labels, "__selected__"]
        element = hv.Points(data, kdims=region_labels, vdims=vdims)
        hover = cls._build_hover_tool(data)
        return element.opts(tools=[hover])

    @classmethod
    def ranges_2d(
        cls, data, region_labels, fields_labels, invert_axes=False, groupby: str | None = None
    ):
        "Vectorizes an nd-overlay of range_2d rectangles."
        kdims = [region_labels[i] for i in (0, 2, 1, 3)]  # LBRT format
        vdims = [*fields_labels, "__selected__"]
        ds = hv.Dataset(data, kdims=kdims, vdims=vdims)
        element = ds.to(hv.Rectangles, groupby=groupby)
        cds_map = dict(zip(region_labels, ("left", "right", "bottom", "top")))
        hover = cls._build_hover_tool(data, cds_map)
        return element.opts(tools=[hover])

    @classmethod
    def ranges_1d(
        cls, data, region_labels, fields_labels, invert_axes=False, groupby: str | None = None
    ):
        """
        Vectorizes an nd-overlay of range_1d rectangles.

        NOTE: Should use VSpans once available!
        """
        vdims = [*fields_labels, "__selected__"]
        ds = hv.Dataset(data, kdims=region_labels, vdims=vdims)
        element = ds.to(hv.VSpans, groupby=groupby)
        hover = cls._build_hover_tool(data)
        return element.opts(tools=[hover])

    @classmethod
    def _build_hover_tool(self, data, cds_map=None) -> HoverTool:
        if cds_map is None:
            cds_map = {}
        tooltips, formatters = [], {}
        for dim in data.columns:
            if dim == "__selected__":
                continue
            cds_name = cds_map.get(dim, dim)
            if data[dim].dtype.kind == "M":
                tooltips.append((dim, f"@{{{cds_name}}}{{%F}}"))
                formatters[f"@{{{cds_name}}}"] = "datetime"
            else:
                tooltips.append((dim, f"@{{{cds_name}}}"))
        return HoverTool(tooltips=tooltips, formatters=formatters)


class AnnotationDisplay(param.Parameterized):
    kdims = param.List(
        default=["x"], bounds=(1, 3), constant=True, doc="Dimensions of the element"
    )

    data = param.DataFrame(doc="Combined dataframe of annotation data", constant=True)

    _count = param.Integer(default=0, precedence=-1)

    def __init__(self, annotator: Annotator, **params) -> None:
        super().__init__(**params)

        self._annotation_count_stream = hv.streams.Params(
            parameterized=self,
            parameters=["_count"],
            transient=True,
        )

        self._selection_info = {}

        self._selection_enabled = True
        self._editable_enabled = True

        transient = False
        self._edit_streams = [
            hv.streams.BoundsXY(transient=transient),
            hv.streams.SingleTap(transient=transient),
            hv.streams.Lasso(transient=transient),
        ]

        self.annotator = weakref.proxy(annotator)
        self.style = weakref.proxy(annotator.style)
        self._update_data()
        self._set_region_format()
        self._element = self._make_empty_element()

    def _set_region_format(self) -> None:
        self.region_format = "-".join([self.annotator.spec[k]["region"] for k in self.kdims])

    def _update_data(self):
        with param.edit_constant(self):
            self.data = self.annotator.get_dataframe(dims=self.kdims)
        if self.annotator.groupby:
            self.annotator.style._groupby = (
                self.annotator.groupby,
                sorted(self.data[self.annotator.groupby].unique()),
            )

    @property
    def element(self):
        return self.overlay()

    @property
    def edit_streams(self) -> dict[str, hv.streams.Stream]:
        edit_streams = {}
        if self.region_format in ("range", "range-range"):
            edit_streams["box_select"] = self._edit_streams[0]
        elif self.region_format in ("point", "point-point"):
            edit_streams["tap"] = self._edit_streams[1]
        return edit_streams

    @property
    def edit_tools(self) -> list[Tool]:
        tools = []
        if self.region_format == "range":
            tools.append(BoxSelectTool(dimensions="width"))
        elif self.region_format == "range-range":
            tools.append(BoxSelectTool())
        elif self.region_format == "point":
            tools.append(BoxSelectTool(dimensions="width"))
        elif self.region_format == "point-point":
            tools.append("tap")
        return tools

    @classmethod
    def _infer_kdim_dtypes(cls, element):
        if not isinstance(element, hv.Element):
            msg = f"Supplied object {element} is not a bare HoloViews Element"
            raise ValueError(msg)
        kdim_dtypes = {}
        for kdim in element.dimensions(selection="key"):
            kdim_dtypes[str(kdim)] = type(element.dimension_values(kdim)[0])
        return kdim_dtypes

    def clear_indicated_region(self):
        "Clear any region currently indicated on the plot by the editor"
        self._edit_streams[0].event(bounds=None)
        self._edit_streams[1].event(x=None, y=None)
        self._edit_streams[2].event(geometry=None)
        self.annotator.clear_regions()

    def _make_empty_element(self) -> hv.Curve | hv.Image:
        El = hv.Curve if len(self.kdims) == 1 else hv.Image
        return El([], kdims=self.kdims).opts(apply_ranges=False)

    @property
    def selection_element(self) -> hv.Element:
        if not hasattr(self, "_selection_element"):
            self._selection_element = self._make_empty_element()
        return self._selection_element

    @property
    def selection_enabled(self) -> bool:
        return self._selection_enabled

    @selection_enabled.setter
    def selection_enabled(self, enabled: bool) -> None:
        self._selection_enabled = enabled
        if not enabled:
            self.clear_indicated_region()

    @property
    def editable_enabled(self) -> bool:
        return self._editable_enabled

    @editable_enabled.setter
    def editable_enabled(self, enabled: bool) -> None:
        self._editable_enabled = enabled

    def _filter_stream_values(self, bounds, x, y, geometry):
        if not self._editable_enabled:
            return (None, None, None, None)
        if self.region_format == "point" and bounds:
            #  x is calculated this way to handle datetime
            x = bounds[0] + (bounds[2] - bounds[0]) / 2
            y = None
            bounds = (x, 0, x, 0)
        elif "range" not in self.region_format:
            bounds = None

        # If selection enabled, tap stream used for selection not for creating point regions
        # if ('point' in self.region_format and self.selection_enabled) or 'point' not in self.region_format:
        if "point" not in self.region_format:
            x, y = None, None

        return bounds, x, y, geometry

    def _make_selection_editor(self) -> hv.DynamicMap:
        def inner(bounds, x, y, geometry):
            bounds, x, y, geometry = self._filter_stream_values(bounds, x, y, geometry)

            info = self.selection_element._get_selection_expr_for_stream_value(
                bounds=bounds, x=x, y=y, geometry=geometry
            )
            (dim_expr, bbox, region_element) = info

            self._selection_info = {
                "dim_expr": dim_expr,
                "bbox": bbox,
                "x": x,
                "y": y,
                "geometry": geometry,
                "region_element": region_element,
            }

            if bbox is not None:
                # self.annotator.set_regions will give recursion error
                self.annotator._set_regions(**bbox)

            kdims = list(self.kdims)
            if self.region_format == "point" and x is not None:
                self.annotator._set_regions(**{kdims[0]: x})
            if None not in [x, y]:
                if len(kdims) == 1:
                    self.annotator._set_regions(**{kdims[0]: x})
                elif len(kdims) == 2:
                    self.annotator._set_regions(**{kdims[0]: x, kdims[1]: y})
                else:
                    msg = "Only 1d and 2d supported for Points"
                    raise ValueError(msg)

            return region_element.opts(*self.style.editor())

        return hv.DynamicMap(inner, streams=self._edit_streams)

    def region_editor(self) -> hv.DynamicMap:
        if not hasattr(self, "_region_editor"):
            self._region_editor = self._make_selection_editor()
        return self._region_editor

    def get_indices_by_position(self, **inputs) -> list[Any]:
        "Return primary key values matching given position in data space"
        if self.annotator.groupby:
            df = self.data[self.data[self.annotator.groupby].isin(self.annotator.visible)]
        else:
            df = self.data

        if df.empty:
            return []

        if "range" in self.region_format:
            iter_mask = (
                (df[f"start[{k}]"] <= v) & (v < df[f"end[{k}]"]) for k, v in inputs.items()
            )
        elif "point" in self.region_format:
            iter_mask = ((df[f"point[{k}]"] - v).abs().argmin() for k, v in inputs.items())
        else:
            msg = f"{self.region_format} not implemented"
            raise NotImplementedError(msg)

        return list(df[reduce(np.logical_and, iter_mask)].index)

    def register_tap_selector(self, element: hv.Element) -> hv.Element:
        def tap_selector(x, y) -> None:  # Tap tool must be enabled on the element
            # Only select the first
            inputs = {str(k): v for k, v in zip(self.kdims, (x, y))}
            indices = self.get_indices_by_position(**inputs)
            if indices:
                self.annotator.select_by_index(indices[0])
            else:
                self.annotator.select_by_index()

        self._tap_stream = hv.streams.Tap(source=element, transient=True)
        self._tap_stream.add_subscriber(tap_selector)
        return element

    def register_double_tap_clear(self, element: hv.Element) -> hv.Element:
        def double_tap_clear(x, y):
            self.clear_indicated_region()

        self._double_tap_stream = hv.streams.DoubleTap(source=element, transient=True)
        self._double_tap_stream.add_subscriber(double_tap_clear)
        return element

    def indicators(self) -> hv.DynamicMap:
        if not hasattr(self, "_indicators"):
            self.register_tap_selector(self._element)
            self.register_double_tap_clear(self._element)

            self._indicators = hv.DynamicMap(
                self.static_indicators,
                streams=[self._annotation_count_stream, self.annotator.param.selected_indices],
            )
        return self._indicators

    def overlay(self, indicators=True, editor=True) -> hv.Overlay:
        layers = []
        active_tools = []
        if "range" in self.region_format or self.region_format == "point":
            active_tools += ["box_select"]
        elif self.region_format == "point-point":
            active_tools += ["tap"]
        layers.append(self._element.opts(tools=self.edit_tools, active_tools=active_tools))

        if indicators:
            layers.append(self.indicators())
        if editor:
            layers.append(self.region_editor())
        return hv.Overlay(layers).collate()

    def static_indicators(self, **events):
        fields_labels = self.annotator.all_fields
        region_labels = [k for k in self.data.columns if k not in fields_labels]

        self.data["__selected__"] = self.data.index.isin(self.annotator.selected_indices)

        indicator_kwargs = {
            "data": self.data,
            "region_labels": region_labels,
            "fields_labels": fields_labels,
            "invert_axes": False,  # Not yet handled
            "groupby": self.annotator.groupby,
        }

        if self.region_format == "range":
            indicator = Indicator.ranges_1d(**indicator_kwargs)
        elif self.region_format == "range-range":
            indicator = Indicator.ranges_2d(**indicator_kwargs)
        elif self.region_format == "point":
            indicator = Indicator.points_1d(**indicator_kwargs)
        elif self.region_format == "point-point":
            indicator = Indicator.points_2d(**indicator_kwargs)
        else:
            msg = f"{self.region_format} not implemented"
            raise NotImplementedError(msg)

        if self.annotator.groupby and self.annotator.visible:
            indicator = indicator.get(self.annotator.visible)
            if indicator is None:
                vis = "', '".join(self.annotator.visible)
                msg = f"Visible dimensions {vis!r} not in spec"
                raise ValueError(msg)

        # Set styling on annotations indicator
        highlight = self.style._indicator_selection
        highlighters = {opt: self._selected_dim_expr(v[0], v[1]) for opt, v in highlight.items()}
        indicator = indicator.opts(*self.style.indicator(**highlighters))

        return indicator.overlay() if self.annotator.groupby else hv.NdOverlay({0: indicator})

    def _selected_dim_expr(self, selected_value, non_selected_value) -> hv.dim:
        if isinstance(non_selected_value, hv.Cycle):
            non_selected_value = non_selected_value.values[0]
        return hv.dim("__selected__").categorize(
            {True: selected_value}, default=non_selected_value
        )

    @property
    def dim_expr(self):
        return self._selection_info["dim_expr"]

    def show_region(self):
        kdims = list(self.kdims)
        region = {k: v for k, v in self.annotator._region.items() if k in self.kdims}

        if not region:
            return

        if self.region_format == "range":
            value = region[kdims[0]]
            bounds = (value[0], 0, value[1], 1)
        elif self.region_format == "range-range":
            bounds = (
                region[kdims[0]][0],
                region[kdims[1]][0],
                region[kdims[0]][1],
                region[kdims[1]][1],
            )
        elif self.region_format == "point":
            value = region[kdims[0]]
            bounds = (value, 0, value, 1)
        else:
            bounds = False

        if bounds:
            self._edit_streams[0].event(bounds=bounds)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pytest

from holonote.annotate import AnnotationTable

if TYPE_CHECKING:
    from holonote.annotate.typing import SpecDict


def _init_table(spec: SpecDict) -> AnnotationTable:
    table = AnnotationTable()
    table.load(primary_key_name="id", fields=["test_description"])
    assert len(table._region_df) == 0, "Should be initialized empty"
    assert tuple(table._region_df.columns) == AnnotationTable.columns
    return table


def _time_range():
    return (pd.Timestamp("2022-06-17"), pd.Timestamp("2022-06-19"))


def test_table_single_kdim() -> None:
    spec = {"TIME": {"type": np.datetime64, "region": "range"}}
    table = _init_table(spec)

    start, end = _time_range()
    regions = {"TIME": (start, end)}
    table.add_annotation(regions, spec=spec, id=100, test_description="A test")

    d = {"region": "range", "dim": "TIME", "value": (start, end), "_id": 100}
    expected = pd.DataFrame([d])
    pd.testing.assert_frame_equal(table._region_df, expected)

    d = {
        "start[TIME]": {100: start},
        "end[TIME]": {100: end},
        "test_description": {100: "A test"},
    }
    expected = pd.DataFrame(d)
    expected.index.name = "id"
    pd.testing.assert_frame_equal(table.get_dataframe(spec=spec), expected)


def test_table_multiple_kdim() -> None:
    spec = {
        "TIME": {"type": np.datetime64, "region": "range"},
        "x": {"type": float, "region": "point"},
    }
    table = _init_table(spec)

    start, end = _time_range()
    regions = {"x": 1, "TIME": (start, end)}
    table.add_annotation(regions, spec=spec, id=100, test_description="A test")

    d = {
        "region": ["point", "range"],
        "dim": ["x", "TIME"],
        "value": [1, (start, end)],
        "_id": [100, 100],
    }
    expected = pd.DataFrame(d)
    pd.testing.assert_frame_equal(table._region_df, expected)

    d = {
        "start[TIME]": {100: start},
        "end[TIME]": {100: end},
        "point[x]": {100: 1.0},
        "test_description": {100: "A test"},
    }
    expected = pd.DataFrame(d)
    expected.index.name = "id"
    pd.testing.assert_frame_equal(table.get_dataframe(spec=spec), expected)


def test_table_multiple_kdim_and_annotations() -> None:
    spec = {
        "TIME": {"type": np.datetime64, "region": "range"},
        "x": {"type": float, "region": "point"},
    }
    table = _init_table(spec)

    start, end = _time_range()
    regions = {"x": 1, "TIME": (start, end)}
    table.add_annotation(regions, spec=spec, id=100, test_description="101")
    regions = {"x": 2}
    table.add_annotation(regions, spec=spec, id=101, test_description="102")

    d = {
        "region": ["point", "range", "point"],
        "dim": ["x", "TIME", "x"],
        "value": [1, (start, end), 2],
        "_id": [100, 100, 101],
    }
    expected = pd.DataFrame(d)
    pd.testing.assert_frame_equal(table._region_df, expected)

    d = {
        "start[TIME]": {100: start, 101: pd.NaT},
        "end[TIME]": {100: end, 101: pd.NaT},
        "point[x]": {100: 1.0, 101: 2.0},
        "test_description": {100: "101", 101: "102"},
    }
    expected = pd.DataFrame(d)
    expected.index.name = "id"
    pd.testing.assert_frame_equal(table.get_dataframe(spec=spec), expected)


def test_only_adding_one_dim_with_multiple_dimensions() -> None:
    spec = {
        "TIME": {"type": np.datetime64, "region": "range"},
        "x": {"type": float, "region": "point"},
    }
    table = _init_table(spec)

    regions = {"x": 1}
    table.add_annotation(regions, spec=spec, id=100, test_description="101")
    regions = {"x": 2}
    table.add_annotation(regions, spec=spec, id=101, test_description="102")

    d = {
        "region": ["point", "point"],
        "dim": ["x", "x"],
        "value": [1, 2],
        "_id": [100, 101],
    }
    expected = pd.DataFrame(d)
    pd.testing.assert_frame_equal(table._region_df, expected)

    d = {
        "start[TIME]": {100: pd.NaT, 101: pd.NaT},
        "end[TIME]": {100: pd.NaT, 101: pd.NaT},
        "point[x]": {100: 1.0, 101: 2.0},
        "test_description": {100: "101", 101: "102"},
    }
    expected = pd.DataFrame(d)
    expected.index.name = "id"
    pd.testing.assert_frame_equal(table.get_dataframe(spec=spec), expected)


def test_nodata_multiple_dimension() -> None:
    spec = {
        "TIME": {"type": np.datetime64, "region": "range"},
        "x": {"type": float, "region": "point"},
    }
    table = _init_table(spec)

    assert table._region_df.empty
    assert tuple(table._region_df.columns) == AnnotationTable.columns

    d = {
        "start[TIME]": [pd.NaT],
        "end[TIME]": [pd.NaT],
        "point[x]": [0.0],
        "test_description": [""],
    }
    expected = pd.DataFrame(d).drop(index=0)
    expected.index.name = "id"
    output = table.get_dataframe(spec=spec)

    assert output.empty
    pd.testing.assert_series_equal(output.dtypes, expected.dtypes)
    pd.testing.assert_index_equal(output.columns, expected.columns)


@pytest.mark.parametrize("dim", ["x", "TIME"])
def test_table_multiple_kdim_and_annotations_with_selected_dims(dim) -> None:
    spec = {
        "TIME": {"type": np.datetime64, "region": "range"},
        "x": {"type": float, "region": "point"},
    }
    table = _init_table(spec)

    start, end = _time_range()
    regions = {"x": 1, "TIME": (start, end)}
    table.add_annotation(regions, spec=spec, id=100, test_description="101")
    regions = {"x": 2}
    table.add_annotation(regions, spec=spec, id=101, test_description="102")

    d = {
        "region": ["point", "range", "point"],
        "dim": ["x", "TIME", "x"],
        "value": [1, (start, end), 2],
        "_id": [100, 100, 101],
    }
    expected = pd.DataFrame(d)
    pd.testing.assert_frame_equal(table._region_df, expected)

    d = {
        "start[TIME]": {100: start, 101: pd.NaT},
        "end[TIME]": {100: end, 101: pd.NaT},
        "point[x]": {100: 1.0, 101: 2.0},
        "test_description": {100: "101", 101: "102"},
    }
    d = {k: v for k, v in d.items() if "[" not in k or dim in k}
    expected = pd.DataFrame(d)
    expected.index.name = "id"
    pd.testing.assert_frame_equal(table.get_dataframe(spec=spec, dims=[dim]), expected)


@pytest.mark.parametrize("dim", ["x", "TIME"])
def test_nodata_multiple_dimension_with_selected_dims(dim) -> None:
    spec = {
        "TIME": {"type": np.datetime64, "region": "range"},
        "x": {"type": float, "region": "point"},
    }
    table = _init_table(spec)

    assert table._region_df.empty

    d = {
        "start[TIME]": [pd.NaT],
        "end[TIME]": [pd.NaT],
        "point[x]": [0.0],
        "test_description": [""],
    }
    d = {k: v for k, v in d.items() if "[" not in k or dim in k}
    expected = pd.DataFrame(d).drop(index=0)
    expected.index.name = "id"
    output = table.get_dataframe(spec=spec, dims=[dim])

    assert output.empty
    pd.testing.assert_series_equal(output.dtypes, expected.dtypes)
    pd.testing.assert_index_equal(output.columns, expected.columns)


def test_table_multiple_kdim_with_wrong_dims() -> None:
    spec = {
        "TIME": {"type": np.datetime64, "region": "range"},
        "x": {"type": float, "region": "point"},
    }
    table = _init_table(spec)

    msg = r"Dimension\(s\) 'bad1' not in the spec"
    with pytest.raises(ValueError, match=msg):
        table.get_dataframe(spec=spec, dims=["bad1"])

    with pytest.raises(ValueError, match=msg):
        table.get_dataframe(spec=spec, dims=["TIME", "bad1"])

    msg = r"Dimension\(s\) 'bad1', 'bad2' not in the spec"
    with pytest.raises(ValueError, match=msg):
        table.get_dataframe(spec=spec, dims=["bad1", "bad2"])



from __future__ import annotations

from collections import namedtuple
from typing import TYPE_CHECKING, Any

import holoviews as hv
import pandas as pd
import param

from .._warnings import warn
from .connector import Connector, SQLiteDB
from .display import AnnotationDisplay, Indicator, Style, _valid_element_opts  # noqa: F401
from .table import AnnotationTable

if TYPE_CHECKING:
    from .typing import SpecDict


event_info = namedtuple("event_info", ["type", "index", "region", "fields"])


class AnnotationEvent(param.Event):
    def _validate(self, val) -> None:
        if isinstance(val, (event_info, bool)):
            return
        msg = "AnnotationEvent must be a event_info or bool"
        raise ValueError(msg)


class AnnotatorInterface(param.Parameterized):
    """
    Baseclass that expresses the Python interface of an Annotator
    without using any holoviews components and without requiring
    display. Most of this API centers around how the Annotator interacts
    with the AnnotationTable.
    """

    selected_indices = param.List(default=[], doc="Indices of selected annotations")

    spec = param.Dict(default={}, doc="Specification of annotation types")

    fields = param.List(default=["description"], doc="List of fields", constant=True)

    static_fields = param.Dict(
        default={},
        constant=True,
        doc="Dictionary with key and value which will be added to each commit",
    )

    default_region = param.Selector(
        default="range", objects=["range", "point"], doc="Default region, if nothing is provided"
    )

    connector = param.ClassSelector(class_=Connector, allow_None=False)

    connector_class = SQLiteDB

    event = AnnotationEvent(
        doc="Event that is triggered when an annotation is created, updated, or deleted"
    )

    def __init__(self, spec, **params):
        if "connector" not in params:
            params["connector"] = self.connector_class()

        spec = self.normalize_spec(spec, default_region=params.get("default_region"))
        super().__init__(spec=spec, **params)
        if set(self.fields) & set(self.static_fields):
            msg = "The values of fields and static_fields must not overlap"
            raise ValueError(msg)
        if self.connector.fields is None:
            self.connector.fields = self.all_fields
        self._region = {}
        self._last_region = None

        self.annotation_table = AnnotationTable()
        self.connector._create_column_schema(self.spec, self.all_fields)
        self.connector._initialize(self.connector.column_schema)
        self.annotation_table.load(self.connector, fields=self.connector.fields, spec=self.spec)

    def __repr__(self):
        settings = [
            f"{name}={type(val).__name__}()"
            if isinstance(val, param.Parameterized)
            else f"{name}={val!r}"
            for name, val in self.param.values().items()
        ]
        return f'{self.__class__.__name__}({", ".join(settings)})'

    @property
    def all_fields(self) -> list:
        """Return a list of all fields including static fields"""
        return [*self.fields, *self.static_fields]

    @classmethod
    def normalize_spec(self, input_spec: dict[str, Any], default_region=None) -> SpecDict:
        """Normalize the spec to conform to SpecDict format

        Accepted input spec formats:
        spec = {
            # Range (two values)
            "A1": (np.float64, "range"),
            "A2": {"type": np.float64, "region": "range"},
            "A3": np.float64,  # Special case
            # Point
            "B1": (np.float64, "point"),
            "B2": {"type": np.float64, "region": "point"},
            # Geometry
            ("C1", "D1"): {"type": np.float64, "region": "geometry"},
            ("C2", "D2"): (np.float64, "geometry"),
        }
        """

        new_spec: SpecDict = {}
        for k, v in input_spec.items():
            if isinstance(v, dict):
                pass
            elif isinstance(v, tuple):
                v = {"type": v[0], "region": v[1]}
            else:
                v = {"type": v, "region": default_region or self.default_region}

            if v["region"] not in ["range", "point", "geometry"]:
                msg = "Region type must be range, point, or geometry."
                raise ValueError(msg)
            if v["region"] == "geometry" and not isinstance(k, tuple):
                msg = "Geometry region dimension must be a tuple."
                raise ValueError(msg)
            new_spec[k] = v

        return new_spec

    @property
    def df(self) -> pd.DataFrame:
        return self.annotation_table.get_dataframe(spec=self.spec)

    def get_dataframe(self, dims) -> pd.DataFrame:
        return self.annotation_table.get_dataframe(spec=self.spec, dims=dims)

    def refresh(self, clear=False):
        "Method to update display state of the annotator and optionally clear stale visual state"

    # Selecting annotations

    def select_by_index(self, *inds) -> None:
        "Set the selection state by the indices i.e. primary key values"
        self.selected_indices = list(inds)

    @property
    def selected_index(self):
        "Convenience property returning a single selected index (the first one) or None"
        return self.selected_indices[0] if len(self.selected_indices) > 0 else None

    @property
    def region(self):
        return self._region

    def set_regions(self, **items):
        self._set_regions(**items)

    def _set_regions(self, **items):
        """Updating regions"""
        # TODO: Validate values based on spec
        for dim, values in items.items():
            if dim not in self.spec:
                msg = f"Dimension {dim} not in spec"
                raise ValueError(msg)
            self._region[dim] = values

    def clear_regions(self):
        self._region = {}
        self._last_region = {}

    def _add_annotation(self, **fields):
        # Primary key specification is optional
        if self.connector.primary_key.field_name not in fields:
            index_val = self.connector.primary_key(self.connector)
            fields[self.connector.primary_key.field_name] = index_val

        # Don't do anything if self.region is an empty dict
        if self.region and self.region != self._last_region:
            self.annotation_table.add_annotation(
                self._region, spec=self.spec, **fields, **self.static_fields
            )
            self._last_region = self._region.copy()
            if "event" in self.param.watchers:
                self.event = event_info(
                    "create", fields[self.connector.primary_key.field_name], self._region, fields
                )

    def add_annotation(self, **fields):
        self._add_annotation(**fields)
        self.clear_regions()

    def update_annotation_region(self, index):
        self.annotation_table.update_annotation_region(self._region, index)
        if "event" in self.param.watchers:
            self.event = event_info("update", index, self._region, None)

    def update_annotation_fields(self, index, **fields):
        self.annotation_table.update_annotation_fields(index, **fields, **self.static_fields)
        if "event" in self.param.watchers:
            self.event = event_info("update", index, None, fields)

    def delete_annotation(self, index):
        try:
            self.annotation_table.delete_annotation(index)
            if "event" in self.param.watchers:
                self.event = event_info("delete", index, None, None)
        except KeyError:
            msg = f"Annotation with index {index!r} does not exist."
            raise ValueError(msg) from None

        if index in self.selected_indices:
            self.selected_indices.remove(index)

    # Defining initial annotations

    def define_annotations(self, data: pd.DataFrame, **kwargs) -> None:
        # Will both set regions and add annotations. Can accept multiple inputs
        # if index is none a new index will be set.
        # if nothing is given it infer the regions and fields from the column header.
        # annotator.define_annotations(
        #     df, TIME=("start", "end"), description="description", index=None
        # )

        # kwargs = dict(A=("start", "end"), description="description")
        # kwargs = dict(A=("start", "end"))
        # kwargs = dict(A=("start", "end"), index=True)

        index = kwargs.pop("index", False)
        field_data = set(self.fields) & set(data.columns)
        f_keys = (set(self.fields) & kwargs.keys()) | field_data
        r_keys = (kwargs.keys() - f_keys) | (set(self.spec) & set(data.columns))
        pk = self.connector.primary_key

        for k, v in kwargs.items():
            if k == v:
                continue
            if v in field_data:
                msg = (
                    f"Input {v!r} has overlapping name with a field or spec. "
                    "This can give weird behavior. Consider renaming the input."
                )
                warn(msg)
        # assert len(set(data.columns) - f_keys - r_keys) == 0

        # Vectorize the conversion?
        for r in data.itertuples(index=index):
            regions = {
                k: tuple([getattr(r, a) for a in kwargs[k]])
                if isinstance(kwargs.get(k, k), tuple)
                else getattr(r, kwargs.get(k, k))
                for k in r_keys
            }
            fields = {k: getattr(r, kwargs.get(k, k)) for k in f_keys}
            if index:
                fields[pk.field_name] = pk.cast(r.Index)

            self._set_regions(**regions)
            self._add_annotation(**fields)

        # See: https://github.com/holoviz/holonote/pull/50
        self.clear_regions()

    # Snapshotting and reverting
    @property
    def has_snapshot(self) -> bool:
        return self.annotation_table.has_snapshot

    def revert_to_snapshot(self) -> None:
        self.annotation_table.revert_to_snapshot()

    def snapshot(self) -> None:
        self.annotation_table.snapshot()

    def commit(self, return_commits=False):
        # self.annotation_table.initialize_table(self.connector)  # Only if not in params
        commits = self.annotation_table.commits(self.connector)
        if return_commits:
            return commits

    def on_event(self, callback) -> None:
        """Register a callback to be called when an annotation event is triggered
        this can be either when an annotation is created, updated, or deleted.

        This is a wrapper around param.bind with watch=True.

        Parameters
        ----------
        callback : function
            function to be called when an annotation event is triggered
        """
        param.bind(callback, self.param.event, watch=True)


class Annotator(AnnotatorInterface):
    """
    An annotator displays the contents of an AnnotationTable and
    provides the means to manipulate view the corresponding contents,
    add new annotations and update existing annotations.
    """

    groupby = param.Selector(default=None, doc="Groupby dimension", allow_refs=True)
    visible = param.ListSelector(
        default=[], doc="Visible dimensions, needs groupby enabled", allow_refs=True
    )
    style = param.ClassSelector(default=Style(), class_=Style, doc="Style parameters")

    def __init__(self, spec: dict, **params):
        """
        The spec argument must be an element or a dictionary of kdim dtypes
        """

        self._displays = {}

        super().__init__(
            spec if isinstance(spec, dict) else self._infer_kdim_dtypes(spec),
            **params,
        )

        self._selection_enabled = True
        self._editable_enabled = True

    @classmethod
    def _infer_kdim_dtypes(self, element: hv.Element) -> dict:
        # Remove?
        return AnnotationDisplay._infer_kdim_dtypes(element)

    def _create_annotation_element(self, element_key: tuple[str, ...]) -> AnnotationDisplay:
        for key in element_key:
            if key not in self.spec:
                msg = f"Dimension {key!r} not in spec"
                raise ValueError(msg)
        return AnnotationDisplay(self, kdims=list(element_key))

    def get_element(self, *kdims: str | hv.Dimension) -> hv.DynamicMap:
        return self.get_display(*kdims).element

    def get_display(self, *kdims: str | hv.Dimension) -> AnnotationDisplay:
        element_key = tuple(map(str, kdims))
        if element_key not in self._displays:
            self._displays[element_key] = self._create_annotation_element(element_key)
        return self._displays[element_key]

    def _get_kdims_from_other_element(self, other):
        if isinstance(other, hv.DynamicMap):
            if other.last is None:
                hv.plotting.util.initialize_dynamic(other)
            other = other.last
        kdims = other.kdims
        if not kdims or kdims == ["Element"]:
            kdims = next(k for el in other.values() if (k := el.kdims))
        kdims = [kdim for kdim in kdims if kdim.name in self.spec]
        if kdims:
            return kdims
        else:
            msg = "No valid kdims found in element"
            raise ValueError(msg)

    def __mul__(self, other: hv.Element | hv.Layout | hv.Overlay | hv.NdOverlay) -> hv.Overlay:
        if isinstance(other, (hv.Overlay, hv.NdOverlay)):
            kdims = self._get_kdims_from_other_element(other)
            opts = other.opts.get().kwargs
            return (other * self.get_element(*kdims)).opts(**opts)
        elif isinstance(other, hv.Layout):
            opts = other.opts.get().kwargs
            to_layout = []
            overlay_opts = _valid_element_opts()["Overlay"]
            for el in other:
                kdims = self._get_kdims_from_other_element(el)
                el_opts = {k: v for k, v in el.opts.get().kwargs.items() if k in overlay_opts}
                to_layout.append((el * self.get_element(*kdims)).opts(**el_opts))
            layout = hv.Layout(to_layout).opts(**opts)
            layout._max_cols = other._max_cols
            return layout
        else:
            kdims = self._get_kdims_from_other_element(other)
            return other * self.get_element(*kdims)

    def __rmul__(self, other: hv.Element) -> hv.Overlay:
        return self.__mul__(other)

    def refresh(self, clear=False) -> None:
        for v in self._displays.values():
            v._update_data()
            hv.streams.Stream.trigger([v._annotation_count_stream])
            if clear:
                v.clear_indicated_region()
            v.show_region()

    def set_annotation_table(self, annotation_table):
        self.select_by_index()
        for v in self._displays.values():
            v.select_by_index()
        self.clear_regions()
        super().set_annotation_table(annotation_table)
        self.refresh(clear=True)

    def clear_edits(self):
        super().clear_edits()
        self.clear_indicated_region()

    def add_annotation(self, **fields):
        super().add_annotation(**fields)
        self.refresh(clear=True)

    def update_annotation_region(self, index):
        super().update_annotation_region(index)
        self.refresh()

    def update_annotation_fields(self, index, **fields):
        super().update_annotation_fields(index, **fields)
        self.refresh()

    def delete_annotation(self, index):
        super().delete_annotation(index)
        self.refresh()

    def delete_annotations(self, *indices):
        if not indices:
            msg = "At least one index must be specified to delete annotations"
            raise ValueError(msg)
        for index in indices:
            super().delete_annotation(index)
        self.refresh()

    def define_annotations(self, data: pd.DataFrame, **kwargs) -> None:
        super().define_annotations(data, **kwargs)
        self.refresh(clear=True)

    def revert_to_snapshot(self):
        super().revert_to_snapshot()
        self.refresh()

    def set_regions(self, **items):
        super().set_regions(**items)
        self.refresh()

    @property
    def selection_enabled(self) -> bool:
        return self._selection_enabled

    @selection_enabled.setter
    def selection_enabled(self, enabled: bool) -> None:
        for v in self._displays.values():
            v.selection_enabled = enabled

        if not enabled:
            self.select_by_index()

    @property
    def editable_enabled(self) -> bool:
        return self._editable_enabled

    @editable_enabled.setter
    def editable_enabled(self, enabled: bool) -> None:
        for v in self._displays.values():
            v.editable_enabled = enabled

    @param.depends("style.param", "groupby", "visible", watch=True)
    def _refresh_style(self) -> None:
        self.refresh()

    @param.depends("fields", watch=True, on_init=True)
    def _set_groupby_objects(self) -> None:
        self.param.groupby.objects = [*self.fields, None]


def annotator_transform(obj):
    if isinstance(obj, AnnotatorInterface):
        return obj.param.event
    return obj


param.reactive.register_reference_transform(annotator_transform)
